# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import Any, Optional

from typing_extensions import override

from oumi.core.datasets import BaseSftDataset
from oumi.core.feature_generators import VisionLanguageConversationFeatureGenerator
from oumi.core.processors.base_processor import BaseProcessor
from oumi.core.tokenizers.base_tokenizer import BaseTokenizer
from oumi.core.types.conversation import (
    Conversation,
)
from oumi.utils.conversation_utils import remove_excessive_images_from_conversation


class VisionLanguageSftDataset(BaseSftDataset, ABC):
    """Abstract dataset for vision-language models.

    This class extends BaseSftDataset to provide functionality specific to
    vision-language tasks. It handles the processing of both image and text data.

    Note:
        This dataset is designed to work with models that can process both
        image and text inputs simultaneously, such as CLIP, BLIP, or other
        multimodal architectures.

    Example:
        >>> from oumi.builders import build_processor, build_tokenizer
        >>> from oumi.core.configs import ModelParams
        >>> from oumi.core.types.conversation import Conversation
        >>> from oumi.core.datasets import VisionLanguageSftDataset
        >>> class MyVisionLanguageSftDataset(VisionLanguageSftDataset):
        ...     def transform_conversation(self, example: dict):
        ...         # Implement the abstract method
        ...         # Convert the raw example into a Conversation object
        ...         pass
        >>> tokenizer = build_tokenizer(
        ...     ModelParams(model_name="Qwen/Qwen2-1.5B-Instruct")
        ... )
        >>> dataset = MyVisionLanguageSftDataset( # doctest: +SKIP
        ...     tokenizer=tokenizer,
        ...     processor_name="openai/clip-vit-base-patch32",
        ...     dataset_name="coco_captions",
        ...     split="train"
        ... )
        >>> sample = next(iter(dataset))  # doctest: +SKIP
        >>> print(sample.keys()) # doctest: +SKIP
    """

    def __init__(
        self,
        *,
        return_conversations: bool = False,
        tokenizer: Optional[BaseTokenizer] = None,
        processor: Optional[BaseProcessor] = None,
        processor_name: Optional[str] = None,
        processor_kwargs: Optional[dict[str, Any]] = None,
        limit: Optional[int] = None,
        trust_remote_code: bool = False,
        max_images: Optional[int] = None,
        **kwargs,
    ) -> None:
        """Initializes a new instance of the VisionLanguageDataset class.

        If `return_conversations` is True, the dataset will return dictionaries
            containing just JSON-encoded `Conversation` objects:
            {"conversation_json": conversation.to_json()}
        Otherwise, the dataset will return dictionaries containing model inputs:
            {"input_ids": ..., "attention_mask": ..., "pixel_values": ...}

        Args:
            tokenizer: A tokenizer for encoding text data.
            processor: An optional processor object for generating features.
            processor_name: The name of the processor to use for feature generation.
            processor_kwargs: A dictionary of processor-specific parameters.
                These parameters are passed to the processor constructor.
                They can override model-specific parameters.
            limit: An optional limit on the number of examples to load.
            trust_remote_code: Whether to trust remote code execution for the processor.
            return_conversations: Whether to return raw `Conversation` objects.
            max_images: The maximum number of images per conversation.
                If the limit is exceeded, the first N images are kept.
                If None or negative, all images are kept intact.
                If 0, all images are dropped.
            **kwargs: Additional keyword arguments to pass to the base class.
        """
        super().__init__(tokenizer=tokenizer, **kwargs)

        self._max_images = max_images

        self._feature_generator = (
            None
            if return_conversations
            else VisionLanguageConversationFeatureGenerator(
                tokenizer=tokenizer,
                processor=processor,
                processor_name=processor_name,
                processor_kwargs=processor_kwargs,
                trust_remote_code=trust_remote_code,
                return_tensors=self._return_tensors,
            )
        )

        if limit is not None:
            # TODO: this should be removed when we switch to datapipes.
            # Right now, we have to iterate over the whole dataset at init time,
            # Which takes way to long.
            self._data = self._data.head(limit)

    @abstractmethod
    def transform_conversation(self, example: dict) -> Conversation:
        """Transforms a raw example into an Oumi Conversation object.

        Args:
            example (dict): A dictionary representing a single conversation example.

        Returns:
            Conversation: A Conversation object representing the conversation.
        """
        raise NotImplementedError

    @override
    def transform(self, sample: dict) -> dict:
        """Transforms an Oumi conversation into a dictionary of inputs for a model.

        Args:
            sample (dict): A dictionary representing a single conversation example.

        Returns:
            dict: A dictionary of inputs for a model.
        """
        conversation = self.transform_conversation(sample)
        if self._max_images is not None and self._max_images >= 0:
            conversation = remove_excessive_images_from_conversation(
                conversation, max_images=self._max_images
            )

        if self._feature_generator is None:
            # This is only compatible with `use_torchdata=True`
            # as HF loaders expect certain keys like `input_ids`.
            conversation_json = conversation.to_json()
            return {"conversation_json": conversation_json}

        return self._feature_generator.transform_conversation(conversation, None)



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path
from typing import Any, Optional, Union

from oumi.core.configs.params.base_params import BaseParams


class ZeRORuntimeStage(str, Enum):
    """DeepSpeed ZeRO optimization stages.

    See DeepSpeed ZeRO documentation: https://www.deepspeed.ai/tutorials/zero/
    """

    ZERO_0 = "0"
    """Disabled ZeRO optimization. Standard data parallelism only."""

    ZERO_1 = "1"
    """Optimizer state sharding only."""

    ZERO_2 = "2"
    """Optimizer state and gradient sharding."""

    ZERO_3 = "3"
    """Full sharding: optimizer state, gradients, and model parameters."""


class DeepSpeedPrecision(str, Enum):
    """Mixed precision options for DeepSpeed."""

    FP16 = "fp16"
    """Half precision floating point."""

    BF16 = "bf16"
    """Brain floating point 16-bit format."""


class DeepSpeedOffloadDevice(str, Enum):
    """Offload device options for DeepSpeed."""

    CPU = "cpu"
    """Offload to CPU memory."""

    NVME = "nvme"
    """Offload to NVMe storage."""


@dataclass
class OffloadConfig:
    """Configuration for DeepSpeed parameter/optimizer offloading."""

    device: DeepSpeedOffloadDevice = DeepSpeedOffloadDevice.CPU
    """Device to offload to."""

    pin_memory: bool = True
    """Whether to use pinned memory for faster transfers."""

    buffer_count: int = 4
    """Number of buffers for overlapping transfers."""

    buffer_count_nvme: int = 2
    """Number of NVMe buffers for overlapping transfers."""

    block_size: int = 1048576
    """Buffer block size for NVMe offloading."""

    queue_depth: int = 8
    """NVMe queue depth."""

    single_submit: bool = False
    """Whether to use single submission for NVMe operations."""

    overlap_events: bool = True
    """Whether to overlap NVMe events."""

    thread_count: int = 1
    """Number of threads for NVMe operations."""


@dataclass
class DeepSpeedParams(BaseParams):
    """Configuration options for DeepSpeed distributed training.

    Based on DeepSpeed configuration schema and best practices from
    LLaMA-Factory implementation.

    For detailed ZeRO configuration options, see:
    https://github.com/microsoft/DeepSpeed/blob/master/deepspeed/runtime/zero/config.py
    """

    enable_deepspeed: bool = False
    """If True, enables DeepSpeed distributed training.

    When False, uses standard PyTorch distributed training (DDP/FSDP).
    When True, initializes DeepSpeed engine with ZeRO optimizations based on zero_stage.
    Allows training larger models using ZeRO optimization stages
    for memory-efficient distributed training.
    """

    deepspeed_config_path: Optional[Union[str, Path]] = None
    """Path to a DeepSpeed JSON configuration file.

    If this parameter is not None, all following fields in this class are ignored.
    Unspecified parameters will use DeepSpeed's internal defaults (not Oumi's defaults).
    """

    zero_stage: ZeRORuntimeStage = ZeRORuntimeStage.ZERO_0
    """ZeRO optimization stage.

    Controls the level of memory optimization:
    - Stage 0: Disabled ZeRO optimization, but DeepSpeed engine still active - DEFAULT
    - Stage 1: Optimizer state sharding
    - Stage 2: Optimizer + gradient sharding
    - Stage 3: Full sharding (optimizer + gradients + parameters)

    Stage 0 is the default to match DeepSpeed behavior. Unlike enable_deepspeed=False,
    Stage 0 still uses DeepSpeed's runtime features without ZeRO memory optimizations.
    Enable higher stages for memory efficiency with large models.
    """

    offload_optimizer: Optional[OffloadConfig] = None
    """Configuration for optimizer state offloading.

    When enabled, optimizer states are offloaded to CPU/NVMe
    to save GPU memory. Available for ZeRO stages 1, 2, and 3.
    """

    offload_param: Optional[OffloadConfig] = None
    """Configuration for parameter offloading.

    When enabled, model parameters are offloaded to CPU/NVMe
    when not actively used. Only available for ZeRO stage 3.
    """

    precision: Optional[DeepSpeedPrecision] = None
    """Mixed precision training mode.

    Options:
    - fp16: Half precision (faster on older GPUs)
    - bf16: Brain float 16 (better numerical stability on newer GPUs)
    - None: Use model's native precision
    """

    # Communication optimization parameters (general)
    overlap_comm: bool = False
    """Whether to overlap communication with computation.

    DeepSpeed default is False. Automatically enabled for ZeRO Stage 3.
    """

    contiguous_gradients: bool = True
    """Whether to ensure gradient memory is contiguous."""

    reduce_bucket_size: Union[int, str] = int(5e8)
    """Bucket size for gradient reduction.

    Can be an integer or "auto" for automatic sizing.
    """

    # ZeRO-1 and ZeRO-2 specific communication parameters
    allgather_bucket_size: int = int(5e8)
    """Bucket size for allgather operations (ZeRO stages 1 and 2)."""

    allgather_partitions: bool = True
    """Enable allgather partitions for ZeRO-2."""

    reduce_scatter: bool = True
    """Enable reduce scatter for ZeRO-2."""

    round_robin_gradients: bool = False
    """Enable round robin gradients (ZeRO stages 1 and 2)."""

    # ZeRO-3 specific parameters
    stage3_prefetch_bucket_size: Union[int, str] = int(5e7)
    """Bucket size for parameter prefetching in ZeRO-3."""

    stage3_param_persistence_threshold: Union[int, str] = int(1e5)
    """Parameter persistence threshold in ZeRO-3."""

    stage3_max_live_parameters: int = int(1e9)
    """Maximum number of live parameters in ZeRO-3."""

    stage3_max_reuse_distance: int = int(1e9)
    """Maximum reuse distance for parameters in ZeRO-3."""

    stage3_gather_16bit_weights_on_model_save: bool = False
    """Whether to gather 16-bit weights during model saving in ZeRO-3.

    DeepSpeed default is False. Set to True if you need full precision weights
    in saved checkpoints.
    """

    sub_group_size: int = int(1e9)
    """Sub-group size for ZeRO-3 parameter sharding."""

    # Training parameters (auto-configured by HuggingFace Transformers)
    train_batch_size: Union[int, str] = "auto"
    """Total training batch size across all GPUs.

    Can be an integer, or "auto" for automatic configuration by HuggingFace.
    When using TRL trainers, this should remain "auto" to allow proper batch size
    management.
    """

    train_micro_batch_size_per_gpu: Union[int, str] = "auto"
    """Micro batch size per GPU.

    Can be an integer, or "auto" for automatic configuration by HuggingFace.
    When using TRL trainers, this should remain "auto" to allow proper batch size
    management.
    """

    gradient_accumulation_steps: Union[int, str] = "auto"
    """Number of gradient accumulation steps.

    Can be an integer, or "auto" for automatic configuration by HuggingFace.
    """

    gradient_clipping: Union[int, str] = "auto"
    """Gradient clipping value.

    Can be an integer, or "auto" for automatic configuration by HuggingFace.
    """

    # Advanced options
    zero_allow_untested_optimizer: bool = True
    """Allow optimizers not explicitly tested with ZeRO."""

    zero_force_ds_cpu_optimizer: bool = True
    """Force DeepSpeed CPU optimizer when CPU offloading is enabled."""

    # Activation checkpointing
    activation_checkpointing: dict[str, Any] = field(default_factory=dict)
    """Configuration for activation checkpointing to save memory.

    DeepSpeed activation checkpointing trades computation for memory by recomputing
    activations during backward pass. Available parameters:

    - partition_activations (bool, default=False): Partition activation checkpoints
      across model parallel GPUs
    - checkpoint_in_cpu (bool, default=False): Move activation checkpoints to CPU.
      Only works when partition_activations=True
    - contiguous_checkpointing (bool, default=False): Copy checkpoints to contiguous
      memory buffer. Requires num_checkpoints to be set
    - num_checkpoints (int, optional): Number of activation checkpoints stored
      during forward propagation. Required for contiguous_checkpointing
    - synchronize (bool, default=False): Perform device synchronization at
      checkpoint boundaries
    - profile (bool, default=False): Log forward/backward time for each checkpoint

    Example: {"partition_activations": True, "checkpoint_in_cpu": True,
              "profile": False}

    For detailed configuration options and API reference, see:
    https://deepspeed.readthedocs.io/en/latest/activation-checkpointing.html
    """

    # Memory optimization
    memory_efficient_linear: bool = False
    """Enable memory-efficient linear layers."""

    # Logging and monitoring
    steps_per_print: int = 10
    """Number of steps between DeepSpeed log prints."""

    wall_clock_breakdown: bool = False
    """Enable detailed wall clock time breakdown logging."""

    def __post_init__(self) -> None:
        """Validate DeepSpeed configuration parameters."""
        # Validate offloading configurations
        if (
            self.offload_param is not None
            and self.zero_stage != ZeRORuntimeStage.ZERO_3
        ):
            raise ValueError(
                "Parameter offloading is only supported with ZeRO stage 3. "
                f"Current stage: {self.zero_stage}"
            )

        if self.offload_optimizer is not None and self.zero_stage not in [
            ZeRORuntimeStage.ZERO_1,
            ZeRORuntimeStage.ZERO_2,
            ZeRORuntimeStage.ZERO_3,
        ]:
            raise ValueError(
                "Optimizer offloading requires ZeRO stage 1, 2, or 3. "
                f"Current stage: {self.zero_stage}"
            )

    def to_deepspeed(self) -> dict[str, Any]:
        """Generate DeepSpeed configuration dictionary.

        Creates a DeepSpeed configuration dict based on the current parameters,
        following the format expected by DeepSpeed's configuration system.

        Returns:
            Dictionary containing DeepSpeed configuration parameters.
        """
        config = {
            "train_batch_size": self.train_batch_size,
            "train_micro_batch_size_per_gpu": self.train_micro_batch_size_per_gpu,
            "gradient_accumulation_steps": self.gradient_accumulation_steps,
            "gradient_clipping": self.gradient_clipping,
            "zero_allow_untested_optimizer": self.zero_allow_untested_optimizer,
            "zero_force_ds_cpu_optimizer": self.zero_force_ds_cpu_optimizer,
            "steps_per_print": self.steps_per_print,
            "wall_clock_breakdown": self.wall_clock_breakdown,
        }

        # Add mixed precision configuration
        if self.precision == DeepSpeedPrecision.FP16:
            config["fp16"] = {
                "enabled": "auto",
                "loss_scale": 0,
                "loss_scale_window": 1000,
                "initial_scale_power": 16,
                "hysteresis": 2,
                "min_loss_scale": 1,
            }
        elif self.precision == DeepSpeedPrecision.BF16:
            config["bf16"] = {"enabled": "auto"}

        # Add ZeRO optimization configuration
        zero_config: dict[str, Any] = {
            "stage": int(self.zero_stage.value),
        }

        # Add stage-specific parameters
        if self.zero_stage in [
            ZeRORuntimeStage.ZERO_0,
            ZeRORuntimeStage.ZERO_1,
            ZeRORuntimeStage.ZERO_2,
        ]:
            zero_config.update(
                {
                    "allgather_partitions": self.allgather_partitions,
                    "allgather_bucket_size": self.allgather_bucket_size,
                    "overlap_comm": self.overlap_comm,
                    "reduce_scatter": self.reduce_scatter,
                    "reduce_bucket_size": self.reduce_bucket_size,
                    "contiguous_gradients": self.contiguous_gradients,
                    "round_robin_gradients": self.round_robin_gradients,
                }
            )

        if self.zero_stage == ZeRORuntimeStage.ZERO_3:
            zero_config.update(
                {
                    "overlap_comm": self.overlap_comm,
                    "contiguous_gradients": self.contiguous_gradients,
                    "sub_group_size": self.sub_group_size,
                    "reduce_bucket_size": self.reduce_bucket_size,
                    "stage3_prefetch_bucket_size": self.stage3_prefetch_bucket_size,
                    "stage3_param_persistence_threshold": (
                        self.stage3_param_persistence_threshold
                    ),
                    "stage3_max_live_parameters": self.stage3_max_live_parameters,
                    "stage3_max_reuse_distance": self.stage3_max_reuse_distance,
                    "stage3_gather_16bit_weights_on_model_save": (
                        self.stage3_gather_16bit_weights_on_model_save
                    ),
                }
            )

        # Add offloading configurations
        if self.offload_optimizer is not None:
            offload_config: dict[str, Any] = {
                "device": self.offload_optimizer.device.value,
                "pin_memory": self.offload_optimizer.pin_memory,
            }
            if self.offload_optimizer.device == DeepSpeedOffloadDevice.NVME:
                offload_config.update(
                    {
                        "buffer_count": self.offload_optimizer.buffer_count_nvme,
                        "block_size": self.offload_optimizer.block_size,
                        "queue_depth": self.offload_optimizer.queue_depth,
                        "single_submit": self.offload_optimizer.single_submit,
                        "overlap_events": self.offload_optimizer.overlap_events,
                        "thread_count": self.offload_optimizer.thread_count,
                    }
                )
            else:
                offload_config["buffer_count"] = self.offload_optimizer.buffer_count

            zero_config["offload_optimizer"] = offload_config

        if self.offload_param is not None:
            offload_config: dict[str, Any] = {
                "device": self.offload_param.device.value,
                "pin_memory": self.offload_param.pin_memory,
            }
            if self.offload_param.device == DeepSpeedOffloadDevice.NVME:
                offload_config.update(
                    {
                        "buffer_count": self.offload_param.buffer_count_nvme,
                        "block_size": self.offload_param.block_size,
                        "queue_depth": self.offload_param.queue_depth,
                        "single_submit": self.offload_param.single_submit,
                        "overlap_events": self.offload_param.overlap_events,
                        "thread_count": self.offload_param.thread_count,
                    }
                )
            else:
                offload_config["buffer_count"] = self.offload_param.buffer_count

            zero_config["offload_param"] = offload_config

        config["zero_optimization"] = zero_config

        # Add activation checkpointing if configured
        if self.activation_checkpointing:
            config["activation_checkpointing"] = self.activation_checkpointing

        return config

    def is_zero3_enabled(self) -> bool:
        """Check if ZeRO stage 3 is enabled."""
        return self.enable_deepspeed and self.zero_stage == ZeRORuntimeStage.ZERO_3

    def requires_param_offloading(self) -> bool:
        """Check if parameter offloading is configured."""
        return self.offload_param is not None

    def requires_optimizer_offloading(self) -> bool:
        """Check if optimizer offloading is configured."""
        return self.offload_optimizer is not None



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any

from typing_extensions import override

from oumi.core.configs import GenerationParams, ModelParams, RemoteParams
from oumi.core.types.conversation import Conversation
from oumi.inference.gcp_inference_engine import (
    _convert_guided_decoding_config_to_api_input,
)
from oumi.inference.remote_inference_engine import RemoteInferenceEngine


class GoogleGeminiInferenceEngine(RemoteInferenceEngine):
    """Engine for running inference against Gemini API."""

    base_url = (
        "https://generativelanguage.googleapis.com/v1beta/openai/chat/completions"
    )
    """The base URL for the Gemini API."""

    api_key_env_varname = "GEMINI_API_KEY"
    """The environment variable name for the Gemini API key."""

    @override
    def _convert_conversation_to_api_input(
        self,
        conversation: Conversation,
        generation_params: GenerationParams,
        model_params: ModelParams,
    ) -> dict[str, Any]:
        """Converts a conversation to an Gemini API input.

        Documentation: https://ai.google.dev/docs

        Args:
            conversation: The conversation to convert.
            generation_params: Parameters for generation during inference.
            model_params: Model parameters to use during inference.

        Returns:
            Dict[str, Any]: A dictionary representing the Gemini input.
        """
        api_input = {
            "model": model_params.model_name,
            "messages": self._get_list_of_message_json_dicts(
                conversation.messages, group_adjacent_same_role_turns=True
            ),
            "max_completion_tokens": generation_params.max_new_tokens,
            "temperature": generation_params.temperature,
            "top_p": generation_params.top_p,
            "n": 1,  # Number of completions to generate for each prompt.
        }

        if generation_params.stop_strings:
            api_input["stop"] = generation_params.stop_strings

        if generation_params.guided_decoding:
            api_input["response_format"] = _convert_guided_decoding_config_to_api_input(
                generation_params.guided_decoding
            )

        return api_input

    @override
    def get_supported_params(self) -> set[str]:
        """Returns a set of supported generation parameters for this engine."""
        return {
            "guided_decoding",
            "max_new_tokens",
            "stop_strings",
            "temperature",
            "top_p",
        }

    @override
    def infer_batch(
        self, conversations: list[Conversation], inference_config: dict[str, Any]
    ) -> str:
        """Run inference on a batch of conversations.

        Args:
            conversations: The batch of conversations to infer on.
            inference_config: The inference configuration.

        Returns:
            str: The batch ID.
        """
        raise NotImplementedError("Batch inference is not supported for Gemini API.")

    @override
    def _default_remote_params(self) -> RemoteParams:
        """Returns the default remote parameters."""
        return RemoteParams(num_workers=2, politeness_policy=60.0)



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Quantization module for Oumi.

This module provides comprehensive model quantization capabilities including
AWQ, BitsAndBytes, and GGUF quantization methods.
"""

from typing import TYPE_CHECKING

from oumi.quantize.awq_quantizer import AwqQuantization
from oumi.quantize.base import BaseQuantization, QuantizationResult
from oumi.quantize.bnb_quantizer import BitsAndBytesQuantization

if TYPE_CHECKING:
    from oumi.core.configs import QuantizationConfig


def quantize(config: "QuantizationConfig") -> QuantizationResult:
    """Main quantization function that routes to appropriate quantizer.

    Args:
        config: Quantization configuration containing method, model parameters,
            and other settings.

    Returns:
        QuantizationResult containing quantization results including file sizes
        and compression ratios.

    Raises:
        ValueError: If quantization method is not supported
        RuntimeError: If quantization fails
    """
    from oumi.core.configs import QuantizationConfig

    if not isinstance(config, QuantizationConfig):
        raise ValueError(f"Expected QuantizationConfig, got {type(config)}")

    # Use builder to create appropriate quantizer
    from oumi.builders.quantizers import build_quantizer

    quantizer = build_quantizer(config.method)
    quantizer.raise_if_requirements_not_met()

    return quantizer.quantize(config)


__all__ = [
    "BaseQuantization",
    "QuantizationResult",
    "AwqQuantization",
    "BitsAndBytesQuantization",
    "quantize",
]



import subprocess
import tempfile
from datetime import datetime, timedelta
from pathlib import Path
from unittest.mock import Mock, call, patch

import pytest

from oumi.core.launcher import JobState, JobStatus
from oumi.launcher.clients.slurm_client import SlurmClient

_CTRL_PATH: str = "-S ~/.ssh/control-%h-%p-%r"
_SACCT_CMD = (
    "sacct --user=user --format='JobId%-30,JobName%30,User%30,State%30,Reason%30' "
    f"-X --starttime {(datetime.now() - timedelta(days=30)).strftime('%Y-%m-%d')}"
)


#
# Fixtures
#
@pytest.fixture
def mock_subprocess_no_init():
    with patch("oumi.launcher.clients.slurm_client.subprocess") as sp:
        yield sp


@pytest.fixture
def mock_subprocess():
    with patch("oumi.launcher.clients.slurm_client.subprocess") as sp:
        sp.TimeoutExpired = subprocess.TimeoutExpired
        mock_child = Mock()
        sp.run.return_value = mock_child
        mock_child.returncode = 0
        yield sp


def _get_test_data(file_name: str) -> str:
    data_path = Path(__file__).parent / "data" / file_name
    with open(data_path) as f:
        return f.read()


def _run_commands_template(commands: list[str]) -> str:
    user = "user"
    ctrl_path = "-S ~/.ssh/control-%h-%p-%r"
    ssh_cmd = f"ssh {ctrl_path} {user}@host  << 'EOF'"
    eof_suffix = "EOF"
    return "\n".join([ssh_cmd, *commands, eof_suffix])


#
# Tests
#
def test_slurm_client_init(mock_subprocess):
    _ = SlurmClient("user", "host", "cluster_name")
    mock_subprocess.run.assert_called_once_with(
        "ssh -S ~/.ssh/control-%h-%p-%r -O check user@host",
        shell=True,
        capture_output=True,
        timeout=10,
    )


def test_slurm_client_submit_job(mock_subprocess):
    mock_run = Mock()
    mock_subprocess.run.return_value = mock_run
    mock_run.stdout = b"2032"
    mock_run.stderr = b"err"
    mock_run.returncode = 0
    client = SlurmClient("user", "host", "cluster_name")
    result = client.submit_job("./job.sh", "work_dir", 2, None)
    mock_subprocess.run.assert_has_calls(
        [
            call(
                "ssh -S ~/.ssh/control-%h-%p-%r -O check user@host",
                shell=True,
                capture_output=True,
                timeout=10,
            ),
            call(
                "ssh -S ~/.ssh/control-%h-%p-%r -O check user@host",
                shell=True,
                capture_output=True,
                timeout=10,
            ),
            call(
                _run_commands_template(
                    [
                        "cd work_dir",
                        "sbatch --nodes=2 --output=$HOME/oumi_slurm_logs/%j.out "
                        "--parsable ./job.sh",
                    ]
                ),
                shell=True,
                capture_output=True,
                timeout=180,
            ),
        ]
    )
    assert result == "2032"


def test_slurm_client_submit_job_name(mock_subprocess):
    mock_run = Mock()
    mock_subprocess.run.return_value = mock_run
    mock_run.stdout = b"2032"
    mock_run.stderr = b"err"
    mock_run.returncode = 0
    client = SlurmClient("user", "host", "cluster_name")
    result = client.submit_job("./job.sh", "work_dir", 2, "somename")
    mock_subprocess.run.assert_has_calls(
        [
            call(
                "ssh -S ~/.ssh/control-%h-%p-%r -O check user@host",
                shell=True,
                capture_output=True,
                timeout=10,
            ),
            call(
                "ssh -S ~/.ssh/control-%h-%p-%r -O check user@host",
                shell=True,
                capture_output=True,
                timeout=10,
            ),
            call(
                _run_commands_template(
                    [
                        "cd work_dir",
                        "sbatch --nodes=2 --job-name=somename "
                        "--output=$HOME/oumi_slurm_logs/%j.out --parsable "
                        "./job.sh",
                    ]
                ),
                shell=True,
                capture_output=True,
                timeout=180,
            ),
        ]
    )
    assert result == "2032"


def test_slurm_client_submit_job_with_all_args(mock_subprocess):
    mock_run = Mock()
    mock_subprocess.run.return_value = mock_run
    mock_run.stdout = b"2032"
    mock_run.stderr = b"err"
    mock_run.returncode = 0
    client = SlurmClient("user", "host", "cluster_name")
    result = client.submit_job(
        "./job.sh",
        "work_dir",
        2,
        name="somename",
        export="NONE",
        account="oumi",
        ntasks=2,
        threads_per_core=1,
        distribution="block:cyclic",
        partition="extended",
        qos="debug",
        stdout_file="~/stdout.txt",
        stderr_file="$HOME/stderr.txt",
        # kwargs
        foo="bar",
    )
    mock_subprocess.run.assert_has_calls(
        [
            call(
                "ssh -S ~/.ssh/control-%h-%p-%r -O check user@host",
                shell=True,
                capture_output=True,
                timeout=10,
            ),
            call(
                "ssh -S ~/.ssh/control-%h-%p-%r -O check user@host",
                shell=True,
                capture_output=True,
                timeout=10,
            ),
            call(
                _run_commands_template(
                    [
                        "cd work_dir",
                        " ".join(
                            [
                                "sbatch",
                                "--nodes=2",
                                "--job-name=somename",
                                "--export=NONE",
                                "--account=oumi",
                                "--ntasks=2",
                                "--threads-per-core=1",
                                "--distribution=block:cyclic",
                                "--partition=extended",
                                "--qos=debug",
                                "--output=~/stdout.txt",
                                "--error=$HOME/stderr.txt",
                                "--foo=bar",
                                "--parsable",
                                "./job.sh",
                            ]
                        ),
                    ]
                ),
                shell=True,
                capture_output=True,
                timeout=180,
            ),
        ]
    )
    assert result == "2032"


def test_slurm_client_submit_job_error(mock_subprocess):
    mock_success_run = Mock()
    mock_success_run.stdout = b"out"
    mock_success_run.stderr = b"err"
    mock_success_run.returncode = 0
    mock_run = Mock()
    mock_subprocess.run.side_effect = [
        mock_success_run,
        mock_success_run,
        mock_run,
    ]
    mock_run.stdout = b"3141592653polaris-pbs-01"
    mock_run.stderr = b"foo"
    mock_run.returncode = 1
    client = SlurmClient("user", "host", "cluster_name")
    with pytest.raises(RuntimeError, match="Failed to submit job. stderr: foo"):
        _ = client.submit_job("./job.sh", "work_dir", 2, None)
    mock_subprocess.run.assert_has_calls(
        [
            call(
                "ssh -S ~/.ssh/control-%h-%p-%r -O check user@host",
                shell=True,
                capture_output=True,
                timeout=10,
            ),
            call(
                "ssh -S ~/.ssh/control-%h-%p-%r -O check user@host",
                shell=True,
                capture_output=True,
                timeout=10,
            ),
            call(
                _run_commands_template(
                    [
                        "cd work_dir",
                        "sbatch --nodes=2 --output=$HOME/oumi_slurm_logs/%j.out "
                        "--parsable ./job.sh",
                    ]
                ),
                shell=True,
                capture_output=True,
                timeout=180,
            ),
        ]
    )


def test_slurm_client_submit_job_retry_auth(mock_subprocess):
    mock_run = Mock()
    mock_subprocess.run.return_value = mock_run
    mock_run.stdout = b"3141592653polaris-pbs-01"
    mock_run.stderr = b"err"
    mock_run.returncode = 0
    client = SlurmClient("user", "host", "cluster_name")
    result = client.submit_job("./job.sh", "work_dir", 2, None)
    mock_subprocess.run.assert_has_calls(
        [
            call(
                "ssh -S ~/.ssh/control-%h-%p-%r -O check user@host",
                shell=True,
                capture_output=True,
                timeout=10,
            ),
            call(
                "ssh -S ~/.ssh/control-%h-%p-%r -O check user@host",
                shell=True,
                capture_output=True,
                timeout=10,
            ),
            call(
                _run_commands_template(
                    [
                        "cd work_dir",
                        "sbatch --nodes=2 --output=$HOME/oumi_slurm_logs/%j.out "
                        "--parsable ./job.sh",
                    ]
                ),
                shell=True,
                capture_output=True,
                timeout=180,
            ),
        ]
    )
    assert result == "3141592653polaris-pbs-01"


def test_slurm_client_list_jobs_success(mock_subprocess):
    mock_run = Mock()
    mock_subprocess.run.return_value = mock_run
    mock_run.stdout = _get_test_data("sacct.txt").encode("utf-8")
    mock_run.stderr = b"foo"
    mock_run.returncode = 0

    client = SlurmClient("user", "host", "cluster_name")
    job_list = client.list_jobs()
    mock_subprocess.run.assert_called_with(
        _run_commands_template([_SACCT_CMD]),
        shell=True,
        capture_output=True,
        timeout=180,
    )
    job_ids = [job.id for job in job_list]
    expected_ids = [
        "6",
        "6.batch",
        "7",
        "7.batch",
    ]
    assert job_ids == expected_ids


def test_slurm_client_list_jobs_first_login_success(mock_subprocess):
    mock_run = Mock()
    mock_subprocess.run.return_value = mock_run
    mock_run.stdout = _get_test_data("sacct_full.txt").encode("utf-8")
    mock_run.stderr = b"foo"
    mock_run.returncode = 0

    client = SlurmClient("user", "host", "cluster_name")
    job_list = client.list_jobs()
    mock_subprocess.run.assert_called_with(
        _run_commands_template([_SACCT_CMD]),
        shell=True,
        capture_output=True,
        timeout=180,
    )
    job_ids = [job.id for job in job_list]
    expected_ids = [
        "5",
        "6",
        "7",
        "8",
        "9",
    ]
    assert job_ids == expected_ids


def test_slurm_client_list_jobs_fails_missing_header(mock_subprocess):
    mock_run = Mock()
    mock_subprocess.run.return_value = mock_run
    data = _get_test_data("sacct_full.txt").encode("utf-8")
    data = b"\n".join(data.split(b"\n")[:-7])
    mock_run.stdout = data
    mock_run.stderr = b"foo"
    mock_run.returncode = 0
    client = SlurmClient("user", "host", "cluster_name")

    with pytest.raises(
        RuntimeError, match="Failed to parse job list. Unexpected format:"
    ):
        client = SlurmClient("user", "host", "cluster_name")
        _ = client.list_jobs()
        mock_subprocess.run.assert_called_with(
            _run_commands_template([_SACCT_CMD]),
            shell=True,
            capture_output=True,
            timeout=180,
        )


def test_slurm_client_list_jobs_handles_empty_string(mock_subprocess):
    mock_run = Mock()
    mock_subprocess.run.return_value = mock_run
    mock_run.stdout = b""
    mock_run.stderr = b"foo"
    mock_run.returncode = 0

    client = SlurmClient("user", "host", "cluster_name")
    job_list = client.list_jobs()
    mock_subprocess.run.assert_called_with(
        _run_commands_template([_SACCT_CMD]),
        shell=True,
        capture_output=True,
        timeout=180,
    )
    job_ids = [job.id for job in job_list]
    expected_ids = []
    assert job_ids == expected_ids


def test_slurm_client_list_jobs_failure(mock_subprocess):
    mock_success_run = Mock()
    mock_success_run.stdout = b"out"
    mock_success_run.stderr = b"err"
    mock_success_run.returncode = 0
    mock_run = Mock()
    mock_subprocess.run.side_effect = [
        mock_success_run,
        mock_success_run,
        mock_success_run,
        mock_run,
    ]
    mock_run.stdout = b""
    mock_run.stderr = b"foo"
    mock_run.returncode = 1

    client = SlurmClient("user", "host", "cluster_name")
    with pytest.raises(RuntimeError, match="Failed to list jobs. stderr: foo"):
        client = SlurmClient("user", "host", "cluster_name")
        _ = client.list_jobs()
    mock_subprocess.run.assert_called_with(
        _run_commands_template([_SACCT_CMD]),
        shell=True,
        capture_output=True,
        timeout=180,
    )


def test_slurm_client_get_job_success(mock_subprocess):
    mock_run = Mock()
    mock_subprocess.run.return_value = mock_run
    mock_run.stdout = _get_test_data("sacct.txt").encode("utf-8")
    mock_run.stderr = b"foo"
    mock_run.returncode = 0

    client = SlurmClient("user", "host", "cluster_name")
    job_status = client.get_job("6")
    mock_subprocess.run.assert_called_with(
        _run_commands_template(
            [
                _SACCT_CMD,
            ]
        ),
        shell=True,
        capture_output=True,
        timeout=180,
    )
    expected_status = JobStatus(
        id="6",
        name="test",
        status="COMPLETED",
        cluster="cluster_name",
        metadata=(
            "JobID                                                 JobName                           User                          State                         Reason\n"  # noqa: E501
            "------------------------------ ------------------------------ ------------------------------ ------------------------------ ------------------------------\n"  # noqa: E501
            "                             6                           test                         taenin                      COMPLETED                           None"  # noqa: E501
        ),
        done=True,
        state=JobState.SUCCEEDED,
    )
    assert job_status == expected_status


def test_slurm_client_get_job_success_one_job(mock_subprocess):
    data = _get_test_data("sacct.txt").encode("utf-8")
    data = b"\n".join(data.split(b"\n")[:3])
    mock_run = Mock()
    mock_subprocess.run.return_value = mock_run
    mock_run.stdout = data
    mock_run.stderr = b"foo"
    mock_run.returncode = 0

    client = SlurmClient("user", "host", "cluster_name")
    job_status = client.get_job("6")
    mock_subprocess.run.assert_called_with(
        _run_commands_template(
            [
                _SACCT_CMD,
            ]
        ),
        shell=True,
        capture_output=True,
        timeout=180,
    )
    expected_status = JobStatus(
        id="6",
        name="test",
        status="COMPLETED",
        cluster="cluster_name",
        metadata=(
            "JobID                                                 JobName                           User                          State                         Reason\n"  # noqa: E501
            "------------------------------ ------------------------------ ------------------------------ ------------------------------ ------------------------------\n"  # noqa: E501
            "                             6                           test                         taenin                      COMPLETED                           None"  # noqa: E501
        ),
        done=True,
        state=JobState.SUCCEEDED,
    )
    assert job_status == expected_status


def test_slurm_client_get_job_not_found(mock_subprocess):
    mock_run = Mock()
    mock_subprocess.run.return_value = mock_run
    mock_run.stdout = _get_test_data("sacct.txt").encode("utf-8")
    mock_run.stderr = b"foo"
    mock_run.returncode = 0
    client = SlurmClient("user", "host", "cluster_name")
    job_status = client.get_job("2017652")
    mock_subprocess.run.assert_called_with(
        _run_commands_template([_SACCT_CMD]),
        shell=True,
        capture_output=True,
        timeout=180,
    )
    assert job_status is None


def test_slurm_client_get_job_not_found_empty(mock_subprocess):
    data = _get_test_data("sacct.txt").encode("utf-8")
    data = b"\n".join(data.split(b"\n")[:2])
    mock_run = Mock()
    mock_subprocess.run.return_value = mock_run
    mock_run.stdout = data
    mock_run.stderr = b"foo"
    mock_run.returncode = 0

    client = SlurmClient("user", "host", "cluster_name")
    job_status = client.get_job("6")
    mock_subprocess.run.assert_called_with(
        _run_commands_template([_SACCT_CMD]),
        shell=True,
        capture_output=True,
        timeout=180,
    )
    assert job_status is None


def test_slurm_client_get_job_failure(mock_subprocess):
    mock_success_run = Mock()
    mock_success_run.stdout = b"out"
    mock_success_run.stderr = b"err"
    mock_success_run.returncode = 0
    mock_run = Mock()
    mock_subprocess.run.side_effect = [
        mock_success_run,
        mock_success_run,
        mock_run,
    ]
    mock_run.stdout = _get_test_data("sacct.txt").encode("utf-8")
    mock_run.stderr = b"foo"
    mock_run.returncode = 1
    client = SlurmClient("user", "host", "cluster_name")
    with pytest.raises(RuntimeError, match="Failed to list jobs. stderr: foo"):
        _ = client.get_job("2017652")
    mock_subprocess.run.assert_called_with(
        _run_commands_template([_SACCT_CMD]),
        shell=True,
        capture_output=True,
        timeout=180,
    )


def test_slurm_client_cancel_success(mock_subprocess):
    mock_run2 = Mock()
    mock_run2.stdout = _get_test_data("sacct.txt").encode("utf-8")
    mock_run2.stderr = b"foo"
    mock_run2.returncode = 0
    mock_subprocess.run.return_value = mock_run2

    client = SlurmClient("user", "host", "cluster_name")
    job_status = client.cancel("7.batch")
    mock_subprocess.run.assert_has_calls(
        [
            call(
                "ssh -S ~/.ssh/control-%h-%p-%r -O check user@host",
                shell=True,
                capture_output=True,
                timeout=10,
            ),
            call(
                "ssh -S ~/.ssh/control-%h-%p-%r -O check user@host",
                shell=True,
                capture_output=True,
                timeout=10,
            ),
            call(
                _run_commands_template(["scancel 7.batch"]),
                shell=True,
                capture_output=True,
                timeout=180,
            ),
            call(
                "ssh -S ~/.ssh/control-%h-%p-%r -O check user@host",
                shell=True,
                capture_output=True,
                timeout=10,
            ),
            call(
                _run_commands_template([_SACCT_CMD]),
                shell=True,
                capture_output=True,
                timeout=180,
            ),
        ]
    )
    expected_status = JobStatus(
        id="7.batch",
        name="batch",
        status="RUNNING",
        cluster="cluster_name",
        metadata=(
            "JobID                                                 JobName                           User                          State                         Reason\n"  # noqa: E501
            "------------------------------ ------------------------------ ------------------------------ ------------------------------ ------------------------------\n"  # noqa: E501
            "                       7.batch                          batch                                                       RUNNING"  # noqa: E501
        ),
        done=False,
        state=JobState.RUNNING,
    )
    assert job_status == expected_status


def test_slurm_client_cancel_scancel_failure(mock_subprocess):
    mock_success_run = Mock()
    mock_success_run.stdout = b"out"
    mock_success_run.stderr = b"err"
    mock_success_run.returncode = 0
    mock_run = Mock()
    mock_subprocess.run.side_effect = [
        mock_success_run,
        mock_success_run,
        mock_run,
    ]
    mock_run.stdout = b""
    mock_run.stderr = b"foo"
    mock_run.returncode = 1
    with pytest.raises(RuntimeError, match="Failed to cancel job. stderr: foo"):
        client = SlurmClient("user", "host", "cluster_name")
        _ = client.cancel("2017652")
    mock_subprocess.run.assert_has_calls(
        [
            call(
                _run_commands_template(["scancel 2017652"]),
                shell=True,
                capture_output=True,
                timeout=180,
            ),
        ]
    )


def test_slurm_client_cancel_sacct_failure(mock_subprocess):
    mock_run1 = Mock()
    mock_run1.stdout = b""
    mock_run1.stderr = b""
    mock_run1.returncode = 0
    mock_run2 = Mock()
    mock_run2.stdout = _get_test_data("sacct.txt").encode("utf-8")
    mock_run2.stderr = b"foo"
    mock_run2.returncode = 1
    mock_subprocess.run.side_effect = [
        mock_run1,
        mock_run1,
        mock_run1,
        mock_run1,
        mock_run2,
    ]
    with pytest.raises(RuntimeError, match="Failed to list jobs. stderr: foo"):
        client = SlurmClient("user", "host", "cluster_name")
        _ = client.cancel("2017652")
    mock_subprocess.run.assert_has_calls(
        [
            call(
                "ssh -S ~/.ssh/control-%h-%p-%r -O check user@host",
                shell=True,
                capture_output=True,
                timeout=10,
            ),
            call(
                "ssh -S ~/.ssh/control-%h-%p-%r -O check user@host",
                shell=True,
                capture_output=True,
                timeout=10,
            ),
            call(
                _run_commands_template(["scancel 2017652"]),
                shell=True,
                capture_output=True,
                timeout=180,
            ),
            call(
                "ssh -S ~/.ssh/control-%h-%p-%r -O check user@host",
                shell=True,
                capture_output=True,
                timeout=10,
            ),
            call(
                _run_commands_template([_SACCT_CMD]),
                shell=True,
                capture_output=True,
                timeout=180,
            ),
        ]
    )


def test_slurm_client_cancel_job_not_found_success(mock_subprocess):
    mock_run2 = Mock()
    mock_run2.stdout = _get_test_data("sacct.txt").encode("utf-8")
    mock_run2.stderr = b"foo"
    mock_run2.returncode = 0
    mock_subprocess.run.return_value = mock_run2
    client = SlurmClient("user", "host", "cluster_name")
    job_status = client.cancel("2017652")
    mock_subprocess.run.assert_has_calls(
        [
            call(
                "ssh -S ~/.ssh/control-%h-%p-%r -O check user@host",
                shell=True,
                capture_output=True,
                timeout=10,
            ),
            call(
                "ssh -S ~/.ssh/control-%h-%p-%r -O check user@host",
                shell=True,
                capture_output=True,
                timeout=10,
            ),
            call(
                _run_commands_template(["scancel 2017652"]),
                shell=True,
                capture_output=True,
                timeout=180,
            ),
            call(
                "ssh -S ~/.ssh/control-%h-%p-%r -O check user@host",
                shell=True,
                capture_output=True,
                timeout=10,
            ),
            call(
                _run_commands_template([_SACCT_CMD]),
                shell=True,
                capture_output=True,
                timeout=180,
            ),
        ]
    )
    assert job_status is None


def test_slurm_client_run_commands_success(mock_subprocess):
    mock_run = Mock()
    mock_subprocess.run.return_value = mock_run
    mock_run.stdout = b"out"
    mock_run.stderr = b"err"
    mock_run.returncode = 0
    commands = [
        "first command",
        "cd second/command",
        "cd third/command",
        "fourth command",
        "cd fifth/command",
        "final command",
    ]
    client = SlurmClient("user", "host", "cluster_name")
    result = client.run_commands(commands)
    mock_subprocess.run.assert_called_with(
        _run_commands_template(commands),
        shell=True,
        capture_output=True,
        timeout=180,
    )
    assert result.exit_code == 0
    assert result.stdout == "out"
    assert result.stderr == "err"


def test_slurm_client_run_commands_success_empty(mock_subprocess):
    mock_run = Mock()
    mock_subprocess.run.return_value = mock_run
    mock_run.stdout = b"out"
    mock_run.stderr = b"err"
    mock_run.returncode = 0
    client = SlurmClient("user", "host", "cluster_name")
    result = client.run_commands([])
    mock_subprocess.run.assert_has_calls(
        [
            call(
                _run_commands_template([]),
                shell=True,
                capture_output=True,
                timeout=180,
            ),
        ]
    )
    assert result.exit_code == 0
    assert result.stdout == "out"
    assert result.stderr == "err"


def test_slurm_client_run_commands_fails(mock_subprocess):
    mock_success_run = Mock()
    mock_success_run.stdout = b"out"
    mock_success_run.stderr = b"err"
    mock_success_run.returncode = 0
    mock_run = Mock()
    mock_subprocess.run.side_effect = [
        mock_success_run,
        mock_success_run,
        mock_run,
    ]
    mock_run.stdout = b"out"
    mock_run.stderr = b"err"
    mock_run.returncode = 1
    client = SlurmClient("user", "host", "cluster_name")
    result = client.run_commands([])
    mock_subprocess.run.assert_called_with(
        _run_commands_template([]),
        shell=True,
        capture_output=True,
        timeout=180,
    )
    assert result.exit_code == 1
    assert result.stdout == "out"
    assert result.stderr == "err"


def test_slurm_client_put_recursive_success(mock_subprocess):
    mock_run = Mock()
    mock_subprocess.run.return_value = mock_run
    mock_run.stdout = b"out"
    mock_run.stderr = b"err"
    mock_run.returncode = 0
    client = SlurmClient("user", "host", "cluster_name")
    client.put_recursive(
        "source",
        "destination",
    )
    mock_subprocess.run.assert_has_calls(
        [
            call(
                'rsync -e "ssh -S ~/.ssh/control-%h-%p-%r" -avz --delete '
                "source user@host:destination",
                shell=True,
                capture_output=True,
                timeout=300,
            ),
        ]
    )


def test_slurm_client_put_recursive_success_gitignore(mock_subprocess):
    with tempfile.TemporaryDirectory() as output_temp_dir:
        with open(Path(output_temp_dir) / ".gitignore", "w") as f:
            f.write("*.txt")
        mock_run = Mock()
        mock_subprocess.run.return_value = mock_run
        mock_run.stdout = b"out"
        mock_run.stderr = b"err"
        mock_run.returncode = 0
        client = SlurmClient("user", "host", "cluster_name")
        client.put_recursive(
            output_temp_dir,
            "destination",
        )
        mock_subprocess.run.assert_has_calls(
            [
                call(
                    'rsync -e "ssh -S ~/.ssh/control-%h-%p-%r" -avz --delete '
                    f"--exclude-from {output_temp_dir}/.gitignore "
                    f"{output_temp_dir} user@host:destination",
                    shell=True,
                    capture_output=True,
                    timeout=300,
                ),
            ]
        )


def test_slurm_client_put_recursive_success_tests(mock_subprocess):
    with tempfile.TemporaryDirectory() as output_temp_dir:
        tests = Path(output_temp_dir) / "tests"
        tests.mkdir()
        with open(tests / "file.txt", "w") as f:
            f.write("*.txt")
        mock_run = Mock()
        mock_subprocess.run.return_value = mock_run
        mock_run.stdout = b"out"
        mock_run.stderr = b"err"
        mock_run.returncode = 0
        client = SlurmClient("user", "host", "cluster_name")
        client.put_recursive(
            output_temp_dir,
            "destination",
        )
        mock_subprocess.run.assert_has_calls(
            [
                call(
                    'rsync -e "ssh -S ~/.ssh/control-%h-%p-%r" -avz --delete '
                    f"--exclude {output_temp_dir}/tests "
                    f"{output_temp_dir} user@host:destination",
                    shell=True,
                    capture_output=True,
                    timeout=300,
                ),
            ]
        )


def test_slurm_client_put_recursive_success_tests_gitignore(mock_subprocess):
    with tempfile.TemporaryDirectory() as output_temp_dir:
        tests = Path(output_temp_dir) / "tests"
        tests.mkdir()
        with open(tests / "file.txt", "w") as f:
            f.write("*.txt")
        with open(Path(output_temp_dir) / ".gitignore", "w") as f:
            f.write("*.txt")
        mock_run = Mock()
        mock_subprocess.run.return_value = mock_run
        mock_run.stdout = b"out"
        mock_run.stderr = b"err"
        mock_run.returncode = 0
        client = SlurmClient("user", "host", "cluster_name")
        client.put_recursive(
            output_temp_dir,
            "destination",
        )
        mock_subprocess.run.assert_has_calls(
            [
                call(
                    'rsync -e "ssh -S ~/.ssh/control-%h-%p-%r" -avz --delete '
                    f"--exclude-from {output_temp_dir}/.gitignore "
                    f"--exclude {output_temp_dir}/tests "
                    f"{output_temp_dir} user@host:destination",
                    shell=True,
                    capture_output=True,
                    timeout=300,
                ),
            ]
        )


def test_slurm_client_put_recursive_failure(mock_subprocess_no_init):
    with tempfile.TemporaryDirectory() as output_temp_dir:
        mock_subprocess_no_init.TimeoutExpired = subprocess.TimeoutExpired
        mock_success_run = Mock()
        mock_success_run.stdout = b"out"
        mock_success_run.stderr = b"err"
        mock_success_run.returncode = 0
        mock_run = Mock()
        mock_subprocess_no_init.run.side_effect = [
            mock_success_run,
            mock_success_run,
            mock_success_run,
            mock_success_run,
            mock_run,
        ]
        mock_run.stdout = b"out"
        mock_run.stderr = b"err"
        mock_run.returncode = 1
        with pytest.raises(RuntimeError, match="Rsync failed. stderr: err"):
            client = SlurmClient("user", "host", "cluster_name")
            client.put_recursive(
                output_temp_dir,
                "destination",
            )
        mock_subprocess_no_init.run.assert_has_calls(
            [
                call(
                    'rsync -e "ssh -S ~/.ssh/control-%h-%p-%r" -avz --delete '
                    f"{output_temp_dir} user@host:destination",
                    shell=True,
                    capture_output=True,
                    timeout=300,
                ),
            ]
        )


def test_slurm_client_put_recursive_timeout(mock_subprocess_no_init):
    with tempfile.TemporaryDirectory() as output_temp_dir:
        mock_subprocess_no_init.TimeoutExpired = subprocess.TimeoutExpired
        mock_success_run = Mock()
        mock_success_run.stdout = b"out"
        mock_success_run.stderr = b"err"
        mock_success_run.returncode = 0
        mock_run = Mock()
        mock_subprocess_no_init.run.side_effect = [
            mock_success_run,
            mock_success_run,
            mock_success_run,
            mock_success_run,
            subprocess.TimeoutExpired("Timeout!", 1),
        ]
        mock_run.stdout = b"out"
        mock_run.stderr = b"err"
        mock_run.returncode = 1
        with pytest.raises(RuntimeError, match="Timeout while running rsync command."):
            client = SlurmClient("user", "host", "cluster_name")
            client.put_recursive(
                output_temp_dir,
                "destination",
            )
        mock_subprocess_no_init.run.assert_has_calls(
            [
                call(
                    'rsync -e "ssh -S ~/.ssh/control-%h-%p-%r" -avz --delete '
                    f"{output_temp_dir} user@host:destination",
                    shell=True,
                    capture_output=True,
                    timeout=300,
                ),
            ]
        )


def test_slurm_client_put_recursive_memory_error(mock_subprocess_no_init):
    with tempfile.TemporaryDirectory() as output_temp_dir:
        mock_subprocess_no_init.TimeoutExpired = subprocess.TimeoutExpired
        mock_success_run = Mock()
        mock_success_run.stdout = b"out"
        mock_success_run.stderr = b"err"
        mock_success_run.returncode = 0
        mock_run = Mock()
        mock_subprocess_no_init.run.side_effect = [
            mock_success_run,
            mock_success_run,
            mock_success_run,
            mock_success_run,
            MemoryError("OOM!"),
        ]
        mock_run.stdout = b"out"
        mock_run.stderr = b"err"
        mock_run.returncode = 1
        with pytest.raises(MemoryError, match="OOM!"):
            client = SlurmClient("user", "host", "cluster_name")
            client.put_recursive(
                output_temp_dir,
                "destination",
            )
        mock_subprocess_no_init.run.assert_has_calls(
            [
                call(
                    'rsync -e "ssh -S ~/.ssh/control-%h-%p-%r" -avz --delete '
                    f"{output_temp_dir} user@host:destination",
                    shell=True,
                    capture_output=True,
                    timeout=300,
                ),
            ]
        )


def test_slurm_client_put_success(mock_subprocess):
    mock_run = Mock()
    mock_subprocess.run.return_value = mock_run
    mock_run.stdout = b"out"
    mock_run.stderr = b"err"
    mock_run.returncode = 0
    client = SlurmClient("user", "host", "cluster_name")
    client.put(
        file_contents="file contents",
        destination="destination/file.txt",
    )
    mock_subprocess.run.assert_has_calls(
        [
            call(
                _run_commands_template(
                    [
                        "mkdir -p destination",
                        "touch destination/file.txt",
                        'cat <<"SCRIPTFILETAG" > destination/file.txt',
                        "file contents",
                        "SCRIPTFILETAG",
                    ]
                ),
                shell=True,
                capture_output=True,
                timeout=180,
            ),
        ]
    )


def test_slurm_client_put_failure(mock_subprocess):
    mock_success_run = Mock()
    mock_success_run.stdout = b"out"
    mock_success_run.stderr = b"err"
    mock_success_run.returncode = 0
    mock_run = Mock()
    mock_subprocess.run.side_effect = [
        mock_success_run,
        mock_success_run,
        mock_run,
    ]
    mock_run.stdout = b"out"
    mock_run.stderr = b"err"
    mock_run.returncode = 1
    with pytest.raises(RuntimeError, match="Failed to write file. stderr: err"):
        client = SlurmClient("user", "host", "cluster_name")
        client.put(
            file_contents="file contents",
            destination="destination/file.txt",
        )
    mock_subprocess.run.assert_has_calls(
        [
            call(
                _run_commands_template(
                    [
                        "mkdir -p destination",
                        "touch destination/file.txt",
                        'cat <<"SCRIPTFILETAG" > destination/file.txt',
                        "file contents",
                        "SCRIPTFILETAG",
                    ]
                ),
                shell=True,
                capture_output=True,
                timeout=180,
            ),
        ]
    )


def test_slurm_client_put_other_exception(mock_subprocess):
    mock_success_run = Mock()
    mock_success_run.stdout = b"out"
    mock_success_run.stderr = b"err"
    mock_success_run.returncode = 0
    mock_subprocess.run.side_effect = [
        mock_success_run,
        mock_success_run,
        ValueError("Dummy test exception!"),
    ]
    with pytest.raises(ValueError, match="Dummy test exception!"):
        client = SlurmClient("user", "host", "cluster_name")
        client.put(
            file_contents="file contents",
            destination="destination/file.txt",
        )
    mock_subprocess.run.assert_has_calls(
        [
            call(
                _run_commands_template(
                    [
                        "mkdir -p destination",
                        "touch destination/file.txt",
                        'cat <<"SCRIPTFILETAG" > destination/file.txt',
                        "file contents",
                        "SCRIPTFILETAG",
                    ]
                ),
                shell=True,
                capture_output=True,
                timeout=180,
            ),
        ]
    )


def test_slurm_client_get_active_users(mock_subprocess):
    mock_run = Mock()
    mock_subprocess.run.return_value = mock_run
    mock_run.stdout = (
        b"control-host-22-matthew\n"
        b"control-host-22-user1\n"
        b"control-host-22-user2\n"
        b"control-host-22-user-with-dash-in-name\n"
    )
    mock_run.stderr = b"foo"
    mock_run.returncode = 0

    active_users = SlurmClient.get_active_users("host")
    mock_subprocess.run.assert_called_with(
        "ls ~/.ssh/ | egrep 'control-host-.*-.*'",
        shell=True,
        capture_output=True,
    )
    assert set(active_users) == {"matthew", "user1", "user2", "user-with-dash-in-name"}


def test_slurm_client_get_active_users_empty(mock_subprocess):
    mock_run = Mock()
    mock_subprocess.run.return_value = mock_run
    mock_run.stdout = b""
    mock_run.stderr = b"foo"
    mock_run.returncode = 0

    active_users = SlurmClient.get_active_users("host")
    mock_subprocess.run.assert_called_with(
        "ls ~/.ssh/ | egrep 'control-host-.*-.*'",
        shell=True,
        capture_output=True,
    )
    assert active_users == []


def test_slurm_client_get_active_users_failure(mock_subprocess):
    mock_run = Mock()
    mock_subprocess.run.return_value = mock_run
    mock_run.stdout = (
        b"control-host-22-matthew\ncontrol-host-22-user1\ncontrol-host-22-user2\n"
    )
    mock_run.stderr = b"foo"
    mock_run.returncode = 1

    active_users = SlurmClient.get_active_users("host")
    mock_subprocess.run.assert_called_with(
        "ls ~/.ssh/ | egrep 'control-host-.*-.*'",
        shell=True,
        capture_output=True,
    )
    assert active_users == []



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oumi.core.datasets import BasePretrainingDataset
from oumi.core.registry import register_dataset


@register_dataset("togethercomputer/RedPajama-Data-V2")
class RedPajamaDataV2Dataset(BasePretrainingDataset):
    """RedPajama V2 Dataset for training large language models.

    This dataset includes over 100B text documents from 84 CommonCrawl snapshots,
    processed using the CCNet pipeline. It contains 30B documents with quality
    signals and 20B deduplicated documents :footcite:`2023_redpajama`.

    The dataset is available in English, German, French, Italian, and Spanish.

    Key Features:
        - Over 100B text documents
        - 30B documents with quality annotations
        - 20B unique documents after deduplication
        - Estimated 50.6T tokens in total (30.4T after deduplication)
        - Quality signals for filtering and analysis
        - Minhash signatures for fuzzy deduplication

    See Also:
        - Hugging Face dataset page:
          https://huggingface.co/datasets/togethercomputer/RedPajama-Data-V2
        - Blog post: https://together.ai/blog/redpajama-data-v2
        - GitHub repo: https://github.com/togethercomputer/RedPajama-Data

    Note:
        - License: Common Crawl Foundation Terms of Use:
          https://commoncrawl.org/terms-of-use
        - Code: Apache 2.0 license

    Citations:
        .. footbibliography::
    """

    default_dataset = "togethercomputer/RedPajama-Data-V2"



# Used in the following published models:
# nyu-visionx/cambrian-phi3-3b
# nyu-visionx/cambrian-8b
# nyu-visionx/cambrian-13b
# nyu-visionx/cambrian-34b

import torch
import torch.nn.functional as F
from transformers import AutoImageProcessor, Dinov2Config, Dinov2Model

from oumi.utils.logging import logger

from .base_encoder import BaseVisionTower


def extract_res_interp(model_name):
    valid_model_prefixes = [
        "facebook/dinov2-small",
        "facebook/dinov2-base",
        "facebook/dinov2-large",
        "facebook/dinov2-giant-imagenet1k-1-layer",
        "facebook/dinov2-giant",
    ]

    for prefix in valid_model_prefixes:
        if model_name.startswith(prefix):
            base_model_name = prefix
            break
    else:
        raise ValueError(f"Unknown vision tower: {model_name}")

    res = None
    interp = None

    parts = model_name[len(base_model_name) :].split("-")
    for part in parts:
        if part.startswith("res"):
            res = int(part[3:])
        elif part.startswith("interp"):
            interp = int(part[6:])

    return base_model_name, res, interp


class DinoVisionTower(BaseVisionTower):
    def __init__(self, vision_tower, args, delay_load=False):
        super().__init__(vision_tower, args, delay_load)

        """try to extract an image resolution and interpolation res from the model name string

        valid model names:
            facebook/dinov2-small
            facebook/dinov2-base
            facebook/dinov2-large
            facebook/dinov2-giant
            facebook/dinov2-giant-imagenet1k-1-layer

        res pattern: <model_name>-res<res>-interp<interp>

        eg: facebook/dinov2-small-res518-interp224
        """

        # extract image resolution from model name
        # if self.vision_tower_name.startswith("facebook/dinov2-"):
        #     parts = self.vision_tower_name.split("-res")
        #     if len(parts) == 2:
        #         self._image_size = int(parts[1])
        #         logger.warning(f"Extracted image size {self._image_size} from passed model name '{self.vision_tower_name}' (using {parts[0]})")
        #         self.vision_tower_name = parts[0]
        #     else:
        #         self._image_size = None  # default is 518px
        base_model_name, res, interp = extract_res_interp(self.vision_tower_name)
        self._vision_tower_name = vision_tower
        self.vision_tower_name = base_model_name
        self._image_size = res
        self._interp_size = interp
        self._patch_size = 14  # default patch size

        if not self.delay_load:
            self.load_model()
        else:
            self.cfg_only = Dinov2Config.from_pretrained(self.vision_tower_name)

    def load_model(self, device_map=None):
        self.vision_tower = Dinov2Model.from_pretrained(self.vision_tower_name)
        """ValueError: Dinov2Model does not support `device_map='auto'`. To implement support, the model class needs to implement the `_no_split_modules` attribute."""
        self.vision_tower._no_split_modules = ["Dinov2SwiGLUFFN"]

        _image_size = self.vision_tower.config.image_size
        if self._image_size is None:
            self._image_size = _image_size
        else:
            logger.warning(
                f"Overriding DinoVisionTower image size of {_image_size} with {self._image_size}"
            )

        # increase shortest edge to prevent edge case crops
        default_shortest_ratio = 8 / 7  # 224/256
        # shortest_edge = int(default_shortest_ratio * self._image_size)
        shortest_edge = self._image_size

        processor = AutoImageProcessor.from_pretrained(
            self.vision_tower_name,
            crop_size=dict(height=self._image_size, width=self._image_size),
            size=dict(shortest_edge=shortest_edge),
        )

        logger.info(f"Dino Vision Processor: {processor}")
        self.image_processor = processor

        # Assign the output channels of the projection convolution as the hidden size
        self._hidden_size = (
            self.vision_tower.embeddings.patch_embeddings.projection.out_channels
        )
        # Assign the first value of the stride of the projection convolution as the patch size
        self._patch_size = (
            self.vision_tower.embeddings.patch_embeddings.projection.stride[0]
        )

        # print(self._hidden_size, self._patch_size)

        self.vision_tower.requires_grad_(self.unfreeze_mm_vision_tower)
        self.is_loaded = True

    @property
    def image_size(self):
        return self._image_size

    def feature_select(self, outputs):
        sequence_output = outputs[
            "last_hidden_state"
        ]  # batch_size, sequence_length, hidden_size

        if self.select_feature == "cls_patch":
            image_features = sequence_output
        elif self.select_feature == "patch":
            image_features = sequence_output[:, 1:]
        elif self.select_feature == "cls":
            image_features = sequence_output[:, 0]
        else:
            raise ValueError(f"Unexpected select feature: {self.select_feature}")
        return image_features

    def interpolate(self, image_features):
        if self._interp_size is None:
            return image_features

        b, num_tokens, dim = image_features.shape

        if num_tokens != self.num_patches:
            target_h = target_w = int(self._interp_size**0.5)
            h = w = int(num_tokens**0.5)

            image_features = image_features.view(b, h, w, dim)
            image_features = image_features.permute(0, 3, 1, 2).contiguous()

            image_features = F.interpolate(
                image_features.to(torch.float32),
                size=(target_h, target_w),
                mode="bilinear",
                align_corners=False,
            ).to(image_features.dtype)

            # Permute the dimensions back to (b, target_h, target_w, dim)
            image_features = image_features.permute(0, 2, 3, 1).contiguous()

            # Flatten the spatial dimensions (target_h, target_w) into a single dimension
            image_features = image_features.flatten(1, 2)

        return image_features

    def _forward(self, images):
        # logger.warning(f"images shape: {images.shape}")
        with torch.set_grad_enabled(self.unfreeze_mm_vision_tower):
            image_forward_outs = self.vision_tower.forward(
                images.to(device=self.device, dtype=self.dtype)
            )
            # logger.warning(f"image_forward_outs shape: {image_forward_outs['last_hidden_state'].shape}")
            image_features = self.feature_select(image_forward_outs).to(images.dtype)
            # logger.warning(f"image_features shape: {image_features.shape}")
            interp_features = self.interpolate(image_features)
            # logger.warning(f"interp_features shape: {interp_features.shape}")
            return interp_features

    @property
    def num_patches_per_side(self):
        return int(self.num_patches**0.5)

    @property
    def num_patches(self):
        if self._interp_size is None:
            return (self._image_size // self._patch_size) ** 2
        else:
            return self._interp_size



import os
from contextlib import contextmanager
from unittest.mock import MagicMock, patch

import pytest

from oumi.core.configs import (
    AutoWrapPolicy,
    BackwardPrefetch,
    FSDPParams,
    ShardingStrategy,
    StateDictType,
    TrainingConfig,
)
from oumi.core.configs.params.training_params import TrainingParams
from oumi.core.distributed import (
    DeviceRankInfo,
    _parse_rank,
    all_gather_object,
    estimate_dataloader_num_workers,
    get_accelerate_env_vars,
    global_leader_first,
    global_leader_only,
    is_local_process_zero,
    is_world_process_zero,
    local_leader_first,
    local_leader_only,
    prepare_accelerate_fsdp_run,
)


#
# Fixtures
#
@pytest.fixture
def mock_torch_distributed():
    with patch("torch.distributed") as mock_dist:
        mock_dist.is_available.return_value = True
        mock_dist.is_initialized.return_value = True
        yield mock_dist


@pytest.fixture
def mock_oumi_barrier():
    with patch("oumi.core.distributed.barrier") as mock_dist:
        yield mock_dist


@pytest.fixture
def mock_torch_barrier():
    with patch("torch.distributed.barrier") as mock:
        yield mock


@pytest.fixture
def mock_device_rank_info():
    with patch("oumi.core.distributed.get_device_rank_info") as mock_info:
        yield mock_info


@pytest.fixture
def mock_work_function():
    mock_work = MagicMock()
    yield mock_work


#
# Utils
#
@contextmanager
def assert_function_called(function, times=1):
    function.reset_mock()
    yield
    assert function.call_count == times


#
# Tests
#
@pytest.mark.parametrize(
    "tested_decorator",
    [local_leader_only, global_leader_only, local_leader_first, global_leader_first],
)
def test_decorators_without_distributed(
    tested_decorator,
    mock_work_function,
    mock_torch_distributed,
    mock_oumi_barrier,
    mock_torch_barrier,
):
    # Disable distributed
    mock_torch_distributed.is_distributed.return_value = False

    # In non-distributed mode, the decorated function should be executed
    # exaclty once without calling any barrier
    with assert_function_called(mock_oumi_barrier, times=1):
        with assert_function_called(mock_torch_barrier, times=0):
            with assert_function_called(mock_work_function, times=1):

                @tested_decorator()
                def test_function():
                    # This should be executed
                    mock_work_function()

                test_function()


@pytest.mark.parametrize(
    "decorator",
    [local_leader_first, global_leader_first],
)
def test_context_managers_without_distributed(
    decorator,
    mock_work_function,
    mock_torch_distributed,
    mock_oumi_barrier,
    mock_torch_barrier,
):
    # Disable distributed
    mock_torch_distributed.is_distributed.return_value = False

    # In non-distributed mode, the decorated function should be executed
    # exaclty once without calling any barrier
    with assert_function_called(mock_oumi_barrier, times=1):
        with assert_function_called(mock_torch_barrier, times=0):
            with assert_function_called(mock_work_function, times=1):
                with decorator():
                    # This should be executed
                    mock_work_function()


@pytest.mark.parametrize(
    "tested_decorator",
    [local_leader_only, global_leader_only],
)
def test_leaders_only_should_do_work(
    tested_decorator, mock_work_function, mock_oumi_barrier, mock_device_rank_info
):
    # Decorated function should be called by local leader
    # Barrier should be called ONCE
    mock_device_rank_info.return_value = DeviceRankInfo(
        world_size=2, rank=0, local_world_size=2, local_rank=0
    )
    assert is_local_process_zero() is True
    assert is_world_process_zero() is True

    with assert_function_called(mock_work_function, times=1):
        with assert_function_called(mock_oumi_barrier):

            @tested_decorator()
            def test_function_should_execute():
                mock_work_function()

            test_function_should_execute()

    # Decorated function should NOT called by other workers
    # Barrier should be called ONCE
    mock_device_rank_info.return_value = DeviceRankInfo(
        world_size=2, rank=1, local_world_size=2, local_rank=1
    )
    assert is_local_process_zero() is False
    assert is_world_process_zero() is False

    with assert_function_called(mock_work_function, times=0):
        with assert_function_called(mock_oumi_barrier):

            @tested_decorator()
            def test_function_should_not_execute():
                mock_work_function.do_work()
                pytest.fail("This should not be executed")

            test_function_should_not_execute()


def test_global_leader_only_should_do_work(
    mock_work_function, mock_oumi_barrier, mock_device_rank_info
):
    # Decorated function should be called by local leader
    # Barrier should be called ONCE
    mock_device_rank_info.return_value = DeviceRankInfo(
        world_size=2, rank=0, local_world_size=2, local_rank=0
    )
    assert is_local_process_zero() is True
    assert is_world_process_zero() is True

    with assert_function_called(mock_work_function, times=1):
        with assert_function_called(mock_oumi_barrier):

            @global_leader_only()
            def test_function_should_execute():
                mock_work_function()

            test_function_should_execute()

    # Decorated function should NOT called by other local leaders
    # Barrier should be called ONCE
    mock_device_rank_info.return_value = DeviceRankInfo(
        world_size=4, rank=3, local_world_size=2, local_rank=0
    )
    assert is_local_process_zero() is True
    assert is_world_process_zero() is False

    with assert_function_called(mock_work_function, times=0):
        with assert_function_called(mock_oumi_barrier):

            @global_leader_only()
            def test_function_should_not_execute():
                mock_work_function.do_work()
                pytest.fail("This should not be executed")

            test_function_should_not_execute()

    # Decorated function should NOT called by other workers
    # Barrier should be called ONCE
    mock_device_rank_info.return_value = DeviceRankInfo(
        world_size=2, rank=1, local_world_size=2, local_rank=1
    )
    assert is_local_process_zero() is False
    assert is_world_process_zero() is False

    with assert_function_called(mock_work_function, times=0):
        with assert_function_called(mock_oumi_barrier):

            @global_leader_only()
            def test_function_should_not_execute():
                mock_work_function.do_work()
                pytest.fail("This should not be executed")

            test_function_should_not_execute()


@pytest.mark.parametrize(
    "tested_decorator",
    [local_leader_first, global_leader_first],
)
def test_decorators_with_distributed(
    tested_decorator,
    mock_work_function,
    mock_oumi_barrier,
    mock_torch_barrier,
    mock_device_rank_info,
):
    # The decorated function should be executed
    # exaclty once and call barrier exactly once
    # for both leaders and workers
    mock_device_rank_info.return_value = DeviceRankInfo(
        world_size=2, rank=0, local_world_size=2, local_rank=0
    )

    with assert_function_called(mock_oumi_barrier, times=1):
        with assert_function_called(mock_work_function, times=1):

            @tested_decorator()
            def test_function():
                # This should be executed
                mock_work_function()

            test_function()

    # Worker node
    mock_device_rank_info.return_value = DeviceRankInfo(
        world_size=2, rank=1, local_world_size=2, local_rank=1
    )

    with assert_function_called(mock_oumi_barrier, times=1):
        with assert_function_called(mock_work_function, times=1):

            @tested_decorator()
            def test_function():
                # This should be executed
                mock_work_function()

            test_function()


def test_estimate_dataloader_num_workers(mock_device_rank_info):
    assert estimate_dataloader_num_workers(gpus_per_node=0, cpu_count=32) == 1
    assert estimate_dataloader_num_workers(gpus_per_node=1, cpu_count=32) == 2
    assert estimate_dataloader_num_workers(gpus_per_node=2, cpu_count=32) == 4
    assert estimate_dataloader_num_workers(gpus_per_node=3, cpu_count=32) == 6
    assert estimate_dataloader_num_workers(gpus_per_node=4, cpu_count=32) == 8
    assert estimate_dataloader_num_workers(gpus_per_node=8, cpu_count=32) == 8

    assert estimate_dataloader_num_workers(gpus_per_node=0, cpu_count=8) == 1
    assert estimate_dataloader_num_workers(gpus_per_node=1, cpu_count=8) == 2
    assert estimate_dataloader_num_workers(gpus_per_node=2, cpu_count=8) == 2
    assert estimate_dataloader_num_workers(gpus_per_node=3, cpu_count=8) == 2
    assert estimate_dataloader_num_workers(gpus_per_node=4, cpu_count=8) == 2
    assert estimate_dataloader_num_workers(gpus_per_node=8, cpu_count=8) == 2

    mock_device_rank_info.return_value = DeviceRankInfo(
        world_size=8, rank=1, local_world_size=4, local_rank=0
    )
    assert estimate_dataloader_num_workers(None, cpu_count=32) == 8
    assert estimate_dataloader_num_workers(None, cpu_count=8) == 2

    mock_device_rank_info.return_value = DeviceRankInfo(
        world_size=2, rank=0, local_world_size=4, local_rank=3
    )
    assert estimate_dataloader_num_workers(None, cpu_count=32) == 8
    assert estimate_dataloader_num_workers(None, cpu_count=8) == 2


def test_all_gather_object_single_gpu(mock_device_rank_info, mock_torch_distributed):
    mock_device_rank_info.return_value = DeviceRankInfo(
        world_size=1, rank=0, local_world_size=1, local_rank=0
    )

    with assert_function_called(mock_device_rank_info, times=2):
        assert all_gather_object({"aa": 12, "bb": 20}) == [{"aa": 12, "bb": 20}]


def test_all_gather_object_multi_gpu(mock_device_rank_info, mock_torch_distributed):
    mock_device_rank_info.return_value = DeviceRankInfo(
        world_size=4, rank=2, local_world_size=2, local_rank=0
    )

    def _all_gather_object_replicate(object_list, obj, group):
        for i in range(len(object_list)):
            object_list[i] = obj

    mock_torch_distributed.all_gather_object = MagicMock(
        side_effect=_all_gather_object_replicate
    )

    with assert_function_called(mock_device_rank_info, times=3):
        assert all_gather_object({"aa": 32, "bb": 40}) == [{"aa": 32, "bb": 40}] * 4

    assert mock_torch_distributed.is_available.call_count == 1
    assert mock_torch_distributed.is_initialized.call_count == 1
    assert mock_torch_distributed.all_gather_object.call_count == 1


def test_get_accelerate_env_vars_default():
    config = TrainingConfig()
    env_vars = get_accelerate_env_vars(config)
    assert env_vars == {
        "ACCELERATE_DYNAMO_BACKEND": "NO",
        "ACCELERATE_DYNAMO_MODE": "default",
        "ACCELERATE_DYNAMO_USE_FULLGRAPH": "False",
        "ACCELERATE_DYNAMO_USE_DYNAMIC": "False",
        "FSDP_USE_ORIG_PARAMS": "true",
        "FSDP_CPU_RAM_EFFICIENT_LOADING": "true",
        "ACCELERATE_USE_FSDP": "false",
        "FSDP_SHARDING_STRATEGY": "FULL_SHARD",
        "FSDP_OFFLOAD_PARAMS": "false",
        "FSDP_BACKWARD_PREFETCH": "BACKWARD_PRE",
        "FSDP_FORWARD_PREFETCH": "false",
        "FSDP_STATE_DICT_TYPE": "FULL_STATE_DICT",
        "FSDP_AUTO_WRAP_POLICY": "NO_WRAP",
        "FSDP_MIN_NUM_PARAMS": "100000",
        "FSDP_SYNC_MODULE_STATES": "true",
        "FSDP_ACTIVATION_CHECKPOINTING": "false",
    }


def test_get_accelerate_env_vars():
    fsdp_params = FSDPParams(
        enable_fsdp=True,
        sharding_strategy=ShardingStrategy.HYBRID_SHARD,
        cpu_offload=True,
        mixed_precision="bf16",
        backward_prefetch=BackwardPrefetch.BACKWARD_POST,
        forward_prefetch=True,
        state_dict_type=StateDictType.SHARDED_STATE_DICT,
        auto_wrap_policy=AutoWrapPolicy.TRANSFORMER_BASED_WRAP,
        transformer_layer_cls="LlamaDecoderLayer",
        sync_module_states=False,
        min_num_params=100000000,
    )
    config = TrainingConfig(
        fsdp=fsdp_params,
        training=TrainingParams(enable_gradient_checkpointing=True, use_peft=True),
    )

    env_vars = get_accelerate_env_vars(config)
    assert env_vars == {
        "ACCELERATE_DYNAMO_BACKEND": "NO",
        "ACCELERATE_DYNAMO_MODE": "default",
        "ACCELERATE_DYNAMO_USE_FULLGRAPH": "False",
        "ACCELERATE_DYNAMO_USE_DYNAMIC": "False",
        "FSDP_USE_ORIG_PARAMS": "false",
        "FSDP_CPU_RAM_EFFICIENT_LOADING": "true",
        "ACCELERATE_USE_FSDP": "true",
        "FSDP_SHARDING_STRATEGY": "HYBRID_SHARD",
        "FSDP_OFFLOAD_PARAMS": "true",
        "ACCELERATE_MIXED_PRECISION": "bf16",
        "FSDP_BACKWARD_PREFETCH": "BACKWARD_POST",
        "FSDP_FORWARD_PREFETCH": "true",
        "FSDP_STATE_DICT_TYPE": "SHARDED_STATE_DICT",
        "FSDP_AUTO_WRAP_POLICY": "TRANSFORMER_BASED_WRAP",
        "FSDP_TRANSFORMER_CLS_TO_WRAP": "LlamaDecoderLayer",
        "FSDP_MIN_NUM_PARAMS": "100000000",
        "FSDP_SYNC_MODULE_STATES": "false",
        "FSDP_ACTIVATION_CHECKPOINTING": "true",
    }


def test_get_accelerate_env_vars_compile_keep_use_orig_params():
    fsdp_params = FSDPParams(
        enable_fsdp=True,
    )
    config = TrainingConfig(
        fsdp=fsdp_params,
        training=TrainingParams(use_peft=True, compile=True),
    )
    env_vars = get_accelerate_env_vars(config)
    assert env_vars == {
        "ACCELERATE_DYNAMO_BACKEND": "NO",
        "ACCELERATE_DYNAMO_MODE": "default",
        "ACCELERATE_DYNAMO_USE_FULLGRAPH": "False",
        "ACCELERATE_DYNAMO_USE_DYNAMIC": "False",
        "FSDP_USE_ORIG_PARAMS": "true",
        "FSDP_CPU_RAM_EFFICIENT_LOADING": "true",
        "ACCELERATE_USE_FSDP": "true",
        "FSDP_SHARDING_STRATEGY": "FULL_SHARD",
        "FSDP_OFFLOAD_PARAMS": "false",
        "FSDP_BACKWARD_PREFETCH": "BACKWARD_PRE",
        "FSDP_FORWARD_PREFETCH": "false",
        "FSDP_STATE_DICT_TYPE": "FULL_STATE_DICT",
        "FSDP_AUTO_WRAP_POLICY": "NO_WRAP",
        "FSDP_MIN_NUM_PARAMS": "100000",
        "FSDP_SYNC_MODULE_STATES": "true",
        "FSDP_ACTIVATION_CHECKPOINTING": "false",
    }


def test_prepare_accelerate_fsdp_run():
    config = TrainingConfig(training=TrainingParams(enable_gradient_checkpointing=True))
    env_vars = prepare_accelerate_fsdp_run(config)
    assert env_vars == {
        "ACCELERATE_DYNAMO_BACKEND": "NO",
        "ACCELERATE_DYNAMO_MODE": "default",
        "ACCELERATE_DYNAMO_USE_FULLGRAPH": "False",
        "ACCELERATE_DYNAMO_USE_DYNAMIC": "False",
        "FSDP_USE_ORIG_PARAMS": "true",
        "FSDP_CPU_RAM_EFFICIENT_LOADING": "true",
        "ACCELERATE_USE_FSDP": "false",
        "FSDP_SHARDING_STRATEGY": "FULL_SHARD",
        "FSDP_OFFLOAD_PARAMS": "false",
        "FSDP_BACKWARD_PREFETCH": "BACKWARD_PRE",
        "FSDP_FORWARD_PREFETCH": "false",
        "FSDP_STATE_DICT_TYPE": "FULL_STATE_DICT",
        "FSDP_AUTO_WRAP_POLICY": "NO_WRAP",
        "FSDP_MIN_NUM_PARAMS": "100000",
        "FSDP_SYNC_MODULE_STATES": "true",
        "FSDP_ACTIVATION_CHECKPOINTING": "true",
    }
    assert not config.training.enable_gradient_checkpointing


def test_prepare_accelerate_fsdp_run_override():
    # Clear all environment variables if set
    expected_env_vars = {
        "ACCELERATE_DYNAMO_BACKEND": "NO",
        "ACCELERATE_DYNAMO_MODE": "default",
        "ACCELERATE_DYNAMO_USE_FULLGRAPH": "False",
        "ACCELERATE_DYNAMO_USE_DYNAMIC": "False",
        "FSDP_USE_ORIG_PARAMS": "true",
        "FSDP_CPU_RAM_EFFICIENT_LOADING": "true",
        "ACCELERATE_USE_FSDP": "false",
        "FSDP_SHARDING_STRATEGY": "FULL_SHARD",
        "FSDP_OFFLOAD_PARAMS": "false",
        "FSDP_BACKWARD_PREFETCH": "BACKWARD_PRE",
        "FSDP_FORWARD_PREFETCH": "false",
        "FSDP_STATE_DICT_TYPE": "FULL_STATE_DICT",
        "FSDP_AUTO_WRAP_POLICY": "NO_WRAP",
        "FSDP_MIN_NUM_PARAMS": "100000",
        "FSDP_SYNC_MODULE_STATES": "true",
        "FSDP_ACTIVATION_CHECKPOINTING": "false",
    }
    for env_var in expected_env_vars:
        if env_var in os.environ:
            del os.environ[env_var]

    # Set env var to override
    os.environ["ACCELERATE_DYNAMO_BACKEND"] = "EXISTING_VALUE"
    config = TrainingConfig(training=TrainingParams())

    # Assert that the env var was overridden and a warning was logged
    with patch("logging.Logger.warning") as mock_warning:
        env_vars = prepare_accelerate_fsdp_run(config)
        mock_warning.assert_called_with(
            "Environment variable `ACCELERATE_DYNAMO_BACKEND` has existing value "
            "`EXISTING_VALUE`, overriding to new value `NO`."
        )
    assert env_vars == expected_env_vars


@pytest.mark.parametrize(
    "rank_input,expected",
    [("1", 1), ("5", 5), ("42", 42), ("100", 100), ("0", 0), ("-1", 0), ("  -1  ", 0)],
)
def test_parse_rank(rank_input, expected):
    """Test that _parse_rank returns correct integer for valid positive rank strings."""
    assert _parse_rank(rank_input) == expected


def test_parse_rank_invalid_non_digit():
    """Test that _parse_rank raises ValueError for non-digit strings."""
    with pytest.raises(ValueError, match=r"Rank must be a number\. Actual: abc\."):
        _parse_rank("abc")

    with pytest.raises(ValueError, match=r"Rank must be a number\. Actual: 1a\."):
        _parse_rank("1a")

    with pytest.raises(ValueError, match=r"Rank must be a number\. Actual: a1\."):
        _parse_rank("a1")


def test_parse_rank_invalid_negative():
    """Test that _parse_rank raises ValueError for negative numbers (except -1)."""
    with pytest.raises(ValueError, match=r"Rank must be a number\. Actual: -2\."):
        _parse_rank("-2")

    with pytest.raises(ValueError, match=r"Rank must be a number\. Actual: -10\."):
        _parse_rank("-10")



import pytest

from oumi.core.configs.internal.supported_models import (
    find_internal_model_config,
    find_internal_model_config_using_model_name,
    find_model_hf_config,
)
from oumi.core.configs.params.model_params import ModelParams


@pytest.mark.parametrize(
    "model_name, trust_remote_code",
    [
        ("llava-hf/llava-1.5-7b-hf", False),
        ("microsoft/Phi-3-vision-128k-instruct", True),
        ("Qwen/Qwen2-VL-2B-Instruct", True),
        ("Salesforce/blip2-opt-2.7b", False),
        # Access is restricted (gated repo):
        # ("meta-llama/Llama-3.2-11B-Vision-Instruct", False),
    ],
)
def test_common_vlm_models(model_name: str, trust_remote_code):
    debug_tag = f"model_name: {model_name} trust_remote_code:{trust_remote_code}"
    assert (
        find_model_hf_config(model_name, trust_remote_code=trust_remote_code)
        is not None
    ), debug_tag

    assert (
        find_internal_model_config_using_model_name(
            model_name, trust_remote_code=trust_remote_code
        )
        is not None
    ), debug_tag

    assert (
        find_internal_model_config(
            ModelParams(model_name=model_name, trust_remote_code=trust_remote_code)
        )
        is not None
    ), debug_tag



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from typing import Union

import pandas as pd
import torch
from torch.utils.data import Dataset
from typing_extensions import override

from oumi.core.datasets.base_dpo_dataset import BaseDpoDataset
from oumi.core.datasets.base_kto_dataset import BaseExperimentalKtoDataset
from oumi.core.datasets.base_pretraining_dataset import BasePretrainingDataset
from oumi.core.datasets.base_sft_dataset import BaseSftDataset
from oumi.core.registry import register_dataset
from oumi.core.types.conversation import Conversation, Message, Role


@register_dataset("debug_classfication")
class DebugClassificationDataset(Dataset):
    def __init__(
        self,
        dataset_size: int = 1000,
        feature_dim: int = 128,
        data_type: str = "float32",
        num_classes: int = 10,
        preprocessing_time_ms: float = 0,
        **kwargs,
    ):
        """Initialize a DebugClassificationDataset.

        This dataset generates random data and labels for debugging purposes.

        Args:
            dataset_size: The size of the dataset.
            feature_dim: The dimension of each feature.
            data_type: The data type of the dataset.
                Supported values are "float32", "float16", and "bfloat16".
            num_classes: The number of classes in the dataset.
            preprocessing_time_ms: The time taken for preprocessing
                in milliseconds.
            **kwargs: Additional keyword arguments.

        Raises:
            ValueError: If the data_type is not supported.
        """
        self.size = dataset_size
        self.feature_dim = feature_dim
        self.data_type = data_type
        self.num_classes = num_classes
        self.preprocessing_time_ms = preprocessing_time_ms

        if self.data_type == "float32":
            dtype = torch.float32
        elif self.data_type == "float16":
            dtype = torch.float16
        elif self.data_type == "bfloat16":
            dtype = torch.bfloat16
        else:
            raise ValueError(f"Unsupported data type: {self.data_type}")

        self.data = torch.randn(self.size, self.feature_dim, dtype=dtype)
        self.labels = torch.randint(0, self.num_classes, (self.size,))

    def __len__(self):
        """Return the size of the dataset."""
        return self.size

    def __getitem__(self, idx):
        """Return the data and label at the given index."""
        if self.preprocessing_time_ms > 0:
            time.sleep(self.preprocessing_time_ms * 1000)
        return {"features": self.data[idx], "labels": self.labels[idx]}


@register_dataset("debug_pretraining")
class DebugPretrainingDataset(BasePretrainingDataset):
    default_dataset = "debug_pretraining"

    def __init__(
        self,
        dataset_size: int = 1000,
        **kwargs,
    ):
        """Initializes a DebugPretrainingDataset.

        Args:
            dataset_size: The size of the dataset.
            **kwargs: Additional keyword arguments.

        """
        self.size = dataset_size

        super().__init__(**kwargs)

    def _load_data(self) -> list[dict]:
        return [{"text": f"This is document number {idx}."} for idx in range(self.size)]


@register_dataset("debug_sft")
class DebugSftDataset(BaseSftDataset):
    default_dataset = "debug_sft"

    def __init__(
        self,
        dataset_size: int = 5,
        **kwargs,
    ):
        """Initializes a DebugSftDataset."""
        self.size = dataset_size

        super().__init__(**kwargs)

    def transform_conversation(self, example: Union[dict, pd.Series]) -> Conversation:
        """Transforms the example into a Conversation object."""
        return Conversation(
            messages=[
                Message(
                    role=Role.USER, content=(example.get("user_message", None) or "")
                ),
                Message(
                    role=Role.ASSISTANT,
                    content=(example.get("assistant_message", None) or ""),
                ),
            ]
        )

    @override
    def _load_data(self) -> pd.DataFrame:
        return pd.DataFrame(
            {
                "user_message": [
                    f"Hello, how are you? (Document number {idx})"
                    for idx in range(self.size)
                ],
                "assistant_message": [
                    f"I'm fine, thank you! (Document number {idx})"
                    for idx in range(self.size)
                ],
            }
        )


@register_dataset("debug_dpo")
class DebugDpoDataset(BaseDpoDataset):
    default_dataset = "debug_dpo"

    def __init__(
        self,
        dataset_size: int = 5,
        **kwargs,
    ):
        """Initializes a DebugSftDataset."""
        self.size = dataset_size

        super().__init__(**kwargs)

    def transform_preference(self, sample: dict) -> dict:
        """Transforms the sample into a preference dict."""
        return {
            "prompt": sample["prompt"],
            "chosen": sample["chosen"],
            "rejected": sample["rejected"],
        }

    @override
    def _load_data(self) -> pd.DataFrame:
        return pd.DataFrame(
            {
                "prompt": [
                    f"Hello, how are you? (Document number {idx})"
                    for idx in range(self.size)
                ],
                "chosen": [
                    f"I'm fine, thank you! (Document number {idx})"
                    for idx in range(self.size)
                ],
                "rejected": [
                    f"fine (Document number {idx})" for idx in range(self.size)
                ],
            }
        )


@register_dataset("debug_kto")
class DebugKtoDataset(BaseExperimentalKtoDataset):
    default_dataset = "debug_kto"

    def __init__(
        self,
        dataset_size: int = 5,
        **kwargs,
    ):
        """Initializes a DebugKtoDataset."""
        self.size = dataset_size

        super().__init__(**kwargs)

    @override
    def _load_data(self) -> pd.DataFrame:
        return pd.DataFrame(
            {
                "prompt": [
                    f"Hello, how are you? (Document number {idx})"
                    for idx in range(self.size)
                ],
                "completion": [
                    f"I'm fine, thank you! (Document number {idx})"
                    for idx in range(self.size)
                ],
                "label": [
                    idx % 2 == 0  # True for even indices, False for odd indices
                    for idx in range(self.size)
                ],
            }
        )



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import io
import os
import re
import signal
import subprocess
import threading
import time
from dataclasses import dataclass
from pathlib import Path
from typing import Callable, Optional, Union

from oumi.core.launcher import JobState, JobStatus
from oumi.utils.logging import logger

_CTRL_PATH = "-S ~/.ssh/control-%h-%p-%r"

_LOG_DIR = "$HOME/oumi_slurm_logs/{job_id}.out"


class _SlurmAuthException(Exception):
    pass


def _check_connection(user: str, slurm_host: str) -> None:
    """Checks if the connection is still open."""
    ssh_cmd = f"ssh {_CTRL_PATH} -O check {user}@{slurm_host}"
    error_msg = ""
    try:
        child = subprocess.run(
            ssh_cmd,
            shell=True,
            capture_output=True,
            timeout=10,
        )
        error_msg = child.stderr.decode("utf-8")
    except subprocess.TimeoutExpired:
        raise _SlurmAuthException("Timeout while checking connection.")
    if child.returncode == 0:
        return
    if error_msg:
        logger.error(f"Error checking connection: {error_msg}")
        error_msg = f" Error: {error_msg}"
    raise _SlurmAuthException("Connection to Slurm host is closed." + error_msg)


def _get_job_state(job_state: str) -> JobState:
    """Gets the JobState from a job state string.

    See https://slurm.schedmd.com/job_state_codes.html for more details.

    Args:
        job_state: The state of the job.

    Returns:
        The JobState.
    """
    failure_states = {
        "BOOT_FAIL",
        "DEADLINE",
        "FAILED",
        "LAUNCH_FAILED",
        "NODE_FAIL",
        "OUT_OF_MEMORY",
        "PREEMPTED",
        "TIMEOUT",
        "SUSPENDED",
        "STOPPED",
    }
    if job_state in failure_states:
        return JobState.FAILED
    elif job_state == "COMPLETED":
        return JobState.SUCCEEDED
    elif job_state == "CANCELLED":
        return JobState.CANCELLED
    elif job_state == "RUNNING":
        return JobState.RUNNING
    return JobState.PENDING


def _is_job_done(job_state: JobState) -> bool:
    """Determines if a job is done based on its state.

    Args:
        job_state: The state of the job.

    Returns:
        True if the job is done, False otherwise.
    """
    return (
        job_state == JobState.SUCCEEDED
        or job_state == JobState.CANCELLED
        or job_state == JobState.FAILED
    )


def _split_status_line(
    line: str, column_lengths: list[int], cluster_name: str, metadata: str
) -> JobStatus:
    """Splits a status line into a JobStatus object.

    The expected order of job fields is:
    0. Job ID
    1. Job Name
    2. User
    3. Job State
    4. Job State Reason

    Sample status report:
    ID      NAME   USER     STATE REASON
    ----- ------ ------ --------- ------
    1     my_job   user COMPLETED    0:0
    2       job2          RUNNING

    Args:
        line: The line to split.
        column_lengths: The lengths in chars of each column in the line.
        cluster_name: The name of the cluster the job is running on.
        metadata: Additional metadata to attach to the job status.

    Returns:
        A JobStatus object.
    """
    if len(column_lengths) != 5:
        raise ValueError(
            f"Expected 5 fields, but found {len(column_lengths)}. Invalid line: {line}."
        )
    fields = []
    # Note: We can't use a simple split() here because empty fields are allowed.
    for i in range(len(column_lengths)):
        start = sum(column_lengths[:i]) + i
        end = start + column_lengths[i]
        fields.append(line[start:end].strip())
    state = _get_job_state(fields[3])
    return JobStatus(
        id=fields[0],
        name=fields[1],
        # JobState can have additional information. The primary state is the first word.
        status=fields[3].split(" ")[0],
        cluster=cluster_name,
        metadata=metadata,
        done=_is_job_done(state),
        state=state,
    )


def _compute_duration_debug_str(start_time: float) -> str:
    duration_sec = time.perf_counter() - start_time
    return f"Duration: {duration_sec:.2f} sec"


@dataclass
class SlurmResponse:
    """A response from Slurm."""

    stdout: str
    stderr: str
    exit_code: int


def retry_auth(user_function: Callable) -> Callable:
    """Decorator to ensure auth is fresh before calling a function."""

    @functools.wraps(user_function)
    def wrapper(self, *args, **kwargs):
        self._refresh_creds()
        return user_function(self, *args, **kwargs)

    return wrapper


class SlurmLogStream(io.TextIOBase):
    """A stream that provides access to job logs.

    This class inherits from io.TextIOBase to provide a file-like interface
    for reading job logs.

    Expected Order of Events:

    1. Initialization: Starts a tail process to monitor the job's stdout file and
       launches a background thread to check job status.
    2. Active Reading: While the job is running, reads lines directly from the
       tail process stdout in real-time.
    3. Job Completion: The background thread detects when the job is done and
       terminates the tail process.
    4. Cleanup: When the stream is closed, resources are cleaned up.

    State Transitions:

    - _proc: None → subprocess.Popen → None (after termination)
    """

    def __init__(
        self,
        cluster_name: str,
        job_id: str,
        client: "SlurmClient",
    ):
        """Initialize the log stream.

        Args:
            cluster_name: The name of the cluster the job was run in.
            job_id: The Slurm job ID whose output file to follow.
            client: The SlurmClient instance.
        """
        self.job_id = job_id
        self._client = client
        self._job_check_thread = None
        self._proc = self._start_tail_process(cluster_name)

        # Progress bar detection patterns
        self._progress_patterns = [
            r"\d+%",  # Percentage patterns like "50%"
            r"\d+/\d+",  # Fraction patterns like "100/200"
            r"\[.*\]",  # Bracket patterns like "[████████░░]"
        ]

    def readline(self) -> str:
        """Read a line from the stream.

        Returns:
            The line read from the stream, or an empty string if the stream is closed.
        """
        if self._proc is None:
            return ""

        # Process is still running, try to read from stdout
        if self._proc and self._proc.stdout:
            line = self._proc.stdout.readline()
            if line:
                if not line.strip():
                    return self.readline()
                if "Starting training" in line:
                    print("starting")
                    return line
                # Convert progress bar lines to use \r instead of \n for updates
                if self._is_progress_line(line):
                    # Remove \n and add \r for in-place progress bar updates
                    return line.rstrip("\n") + "\r"
                return line
        return ""

    def _is_progress_line(self, line: str) -> bool:
        """Check if a line looks like a progress bar update.

        Args:
            line: The line to check.

        Returns:
            True if the line appears to be a progress bar update.
        """
        # Look for multiple progress indicators in the same line
        pattern_matches = 0
        for pattern in self._progress_patterns:
            if re.search(pattern, line):
                pattern_matches += 1

        # If we have multiple progress indicators, it's likely a progress bar
        if pattern_matches >= 2:
            return True

        return False

    def close(self) -> None:
        """Close the stream and clean up resources."""
        if self._proc:
            try:
                os.killpg(self._proc.pid, signal.SIGINT)
            except ProcessLookupError:
                pass  # process already gone and can happen due to race conditions
            except Exception as e:
                logger.exception(f"Failed to kill process: {e}")
            finally:
                self._proc = None

    def __enter__(self):
        """Context manager entry."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Context manager exit."""
        self.close()

    def _start_tail_process(self, cluster_name: str) -> subprocess.Popen:
        """Starts a tail process for the specified job.

        This is an internal method that starts the SSH tail process and returns
        the subprocess object for the LogStream to read from.

        Args:
            cluster_name: The name of the cluster the job was run in.

        Returns:
            A subprocess.Popen object that can be read from.
        """
        max_attempts = 6
        base_delay = 5
        max_delay = 20

        log_path = Path(_LOG_DIR.format(job_id=self.job_id))
        check_cmd = f"ssh {_CTRL_PATH} {cluster_name} test -f {log_path}"
        # Wait for log file to appear
        for attempt in range(max_attempts):
            preflight = subprocess.run(
                check_cmd,
                shell=True,
                capture_output=True,
                timeout=30,
            )
            if preflight.returncode == 0:
                break

            if attempt < max_attempts - 1:
                delay = min(base_delay * (2**attempt), max_delay)
                time.sleep(delay)
            else:
                raise FileNotFoundError(
                    f"Log file not found after {max_attempts} attempts: {log_path}. "
                    "The job may not have started."
                )
        logger.info(f"Log location: {log_path}")

        tail_cmd = f"ssh {_CTRL_PATH} {cluster_name} tail -n +1 -F {log_path}"

        proc = subprocess.Popen(
            tail_cmd,
            shell=True,
            stdin=subprocess.DEVNULL,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            start_new_session=True,
            text=True,
            bufsize=1,
        )

        self._start_job_checking(proc)

        return proc

    def _start_job_checking(self, proc: subprocess.Popen):
        """Start background thread to check job status."""

        def check_job_status():
            while True:
                try:
                    job = self._client.get_job(self.job_id)
                    if job is None or job.done:
                        proc.terminate()
                        proc.wait()
                        break
                    time.sleep(2)
                except Exception:
                    break

        self._job_check_thread = threading.Thread(target=check_job_status, daemon=True)
        self._job_check_thread.start()


class SlurmClient:
    """A client for communicating with a Slurm cluster."""

    def __init__(self, user: str, slurm_host: str, cluster_name: str):
        """Initializes a new instance of the SlurmClient class.

        Args:
            user: The user to act as.
            slurm_host: The Slurm Head Node to connect to.
            cluster_name: The name of the cluster this client communicates with.
        """
        self._user = user
        self._slurm_host = slurm_host
        self._cluster_name = cluster_name
        self._refresh_creds()

    def _refresh_creds(self):
        """Refreshes the credentials for the client."""
        try:
            _check_connection(self._user, self._slurm_host)
            # We have fresh credentials, so we return early.
            return
        except _SlurmAuthException:
            logger.warning("No connection found. Establishing a new SSH tunnel...")
        ssh_cmd = (
            f'ssh -f -N -M {_CTRL_PATH} -o "ControlPersist 4h" '
            f"{self._user}@{self._slurm_host}"
        )
        child = subprocess.run(
            ssh_cmd,
            shell=True,
            capture_output=True,
            timeout=180,  # time in seconds
        )
        if child.returncode != 0:
            output = child.stderr.decode("utf-8")
            logger.error(f"Credential error: {output}")
            raise RuntimeError(
                "Failed to refresh Slurm credentials "
                f"for {self._user}@{self._slurm_host}."
            )
        return SlurmResponse(
            stdout=child.stdout.decode("utf-8"),
            stderr=child.stderr.decode("utf-8"),
            exit_code=child.returncode,
        )

    @staticmethod
    def get_active_users(slurm_host: str) -> list[str]:
        """Gets the list of users with an open SSH tunnel to a Slurm cluster.

        Returns:
            A list of users.
        """
        # List all active users with an open SSH tunnel to the Slurm head node.
        command = f"ls ~/.ssh/ | egrep 'control-{slurm_host}-.*-.*'"
        result = subprocess.run(command, shell=True, capture_output=True)
        if result.returncode != 0:
            return []
        # Sample Pattern:
        # control-HOSTNAME-22-taenin
        ssh_tunnel_pattern = rf"control-{slurm_host}-[^-]*-(.*)"
        lines = result.stdout.decode("utf-8").strip().split("\n")
        users = set()
        for line in lines:
            match = re.match(ssh_tunnel_pattern, line.strip())
            if match:
                users.add(match.group(1))
        return list(users)

    def _parse_job_id(self, sbatch_output: str) -> str:
        """Parses the job ID from the result of sbatch.

        From the Slurm MAN page:
        Outputs only the job id number and the cluster name if present.
        The values are separated by a semicolon. Errors will still be displayed.

        Args:
            sbatch_output: The result of sbatch

        Returns:
            The job ID.
        """
        split_job = sbatch_output.strip().split(";")
        if len(split_job) > 2:
            raise ValueError(f"Unexpected output from sbatch: {sbatch_output}")
        return split_job[0]

    @retry_auth
    def run_commands(self, commands: list[str]) -> SlurmResponse:
        """Runs the provided commands in a single SSH command.

        Args:
            commands: The commands to run.
        """
        ssh_cmd = f"ssh {_CTRL_PATH} {self._user}@{self._slurm_host}  << 'EOF'"
        eof_suffix = "EOF"
        new_cmd = "\n".join([ssh_cmd, *commands, eof_suffix])
        start_time: float = time.perf_counter()
        try:
            logger.debug(f"Running commands:\n{new_cmd}")
            child = subprocess.run(
                new_cmd,
                shell=True,
                capture_output=True,
                timeout=180,  # time in seconds
            )
            duration_str = _compute_duration_debug_str(start_time)
            return_code = int(child.returncode)
            stdout_str = child.stdout.decode("utf-8")
            stderr_str = child.stderr.decode("utf-8")
            if return_code == 0:
                logger.debug(f"Commands successfully finished! {duration_str}")
            else:
                logger.error(
                    f"Commands failed with code: {return_code}! {duration_str}"
                    f"\n\nSTDERR: {stderr_str}"
                    f"\n\nSTDOUT: {stdout_str}"
                )
            return SlurmResponse(
                stdout=stdout_str,
                stderr=stderr_str,
                exit_code=return_code,
            )
        except subprocess.TimeoutExpired:
            duration_str = _compute_duration_debug_str(start_time)
            logger.exception(f"Commands timed out ({duration_str})! {new_cmd}")
            return SlurmResponse(
                stdout="",
                stderr=f"Timeout while running command: {new_cmd}",
                exit_code=1,
            )
        except Exception:
            duration_str = _compute_duration_debug_str(start_time)
            logger.exception(f"Command failed ({duration_str})! {new_cmd}")
            raise

    def submit_job(
        self,
        job_path: str,
        working_dir: str,
        node_count: int,
        name: Optional[str],
        *,
        export: Optional[Union[str, list[str]]] = None,
        account: Optional[str] = None,
        ntasks: Optional[int] = None,
        threads_per_core: Optional[int] = None,
        distribution: Optional[str] = None,
        partition: Optional[str] = None,
        qos: Optional[str] = None,
        stdout_file: Optional[str] = _LOG_DIR.format(job_id="%j"),
        stderr_file: Optional[str] = None,
        **kwargs,
    ) -> str:
        """Submits the specified job script to Slurm.

        Args:
            job_path: The path to the job script to submit.
            working_dir: The working directory to submit the job from.
            node_count: The number of nodes to use for the job.
            name: The name of the job (optional).
            export: Environment variables to export.
                Special values: "NONE" nothing to export, "ALL" export all env vars.
            account: Charge job to specified account/project.
            ntasks: Total number of tasks to run.
            threads_per_core: Number of threads per core to allocate
                e.g., 1 to allow only one thread per core,
                or 2 to make use of hyper-threading.
            distribution: Distribution method for processes to nodes
                (type = block|cyclic|arbitrary)
            partition: Partition (aka queue) requested.
            qos: QoS (aka the queue on Perlmutter) requested.
            stdout_file: The file to write the stdout to.
            stderr_file: The file to write the stderr to.
            kwargs: Additional flags to pass to sbatch. Hyphens in the flag name are
                replaced with underscores. For example, `foo_bar=baz` as a kwarg will
                add "--foo-bar=baz" to the sbatch command.

        Returns:
            The ID of the submitted job.
        """
        cmd_parts = ["sbatch", f"--nodes={node_count}"]
        slurm_flags: dict[str, str] = {}
        if name:
            slurm_flags["job-name"] = name
        if export is not None:
            export_str = "NONE"
            if isinstance(export, list):
                if len(export) > 0:
                    export_str = ",".join(export)
            else:
                export_str = str(export)
            slurm_flags["export"] = export_str
        if account:
            slurm_flags["account"] = account
        if ntasks is not None:
            slurm_flags["ntasks"] = str(ntasks)
        if threads_per_core is not None:
            slurm_flags["threads-per-core"] = str(threads_per_core)
        if distribution:
            slurm_flags["distribution"] = distribution
        if partition:
            slurm_flags["partition"] = partition
        if qos:
            slurm_flags["qos"] = qos
        if stdout_file:
            slurm_flags["output"] = stdout_file
        if stderr_file:
            slurm_flags["error"] = stderr_file

        # Add kwargs to slurm_flags
        for flag, value in kwargs.items():
            flag = flag.replace("_", "-")
            if flag in slurm_flags:
                logger.warning(
                    f"Flag {flag} already set to {slurm_flags[flag]}. "
                    f"Overwriting with {value}."
                )
            slurm_flags[flag] = str(value)

        # Add flags to command parts
        for flag, value in slurm_flags.items():
            cmd_parts.append(f"--{flag}={value}")

        cmd_parts.append("--parsable")
        cmd_parts.append(job_path)
        sbatch_cmd = " ".join(cmd_parts)
        logger.debug(f"Executing SBATCH command: {sbatch_cmd}")
        result = self.run_commands([f"cd {working_dir}", sbatch_cmd])
        if result.exit_code != 0:
            raise RuntimeError(f"Failed to submit job. stderr: {result.stderr}")
        return self._parse_job_id(result.stdout.strip())

    def list_jobs(self) -> list[JobStatus]:
        """Lists all jobs for the current user.

        Returns:
            A list of JobStatus.
        """
        response_format = "JobId%-30,JobName%30,User%30,State%30,Reason%30"
        # Get current date and subtract one month.
        # Otherwise completed jobs older than ~24 hours may not be listed.

        from datetime import datetime, timedelta

        current_date = datetime.now()
        one_month_ago = current_date - timedelta(days=30)
        start_date = one_month_ago.strftime("%Y-%m-%d")

        command = (
            f"sacct --user={self._user} --format='{response_format}' -X "
            f"--starttime {start_date}"
        )
        result = self.run_commands([command])
        if result.exit_code != 0:
            raise RuntimeError(f"Failed to list jobs. stderr: {result.stderr}")
        # Parse STDOUT to retrieve job statuses.
        lines = result.stdout.strip().split("\n")
        jobs = []
        if len(lines) < 2:
            return jobs
        # Look for a line starting in JobID followed by a line starting with "--".
        start_idx = -1
        for idx in range(len(lines) - 1):
            if lines[idx].startswith("JobID") and lines[idx + 1].startswith("--"):
                start_idx = idx
                break
        if start_idx == -1:
            raise RuntimeError(
                f"Failed to parse job list. Unexpected format: {result.stdout}"
            )
        lines = lines[start_idx:]
        # The first two lines are metadata headers.
        # The top line is composed of column titles.
        # The second line is composed of ---- characters, each the length of a column.
        metadata_headers = lines[:2]
        column_lengths = [len(col) for col in lines[1].strip().split(" ")]
        job_lines = lines[2:]
        for line in job_lines:
            job_metadata = "\n".join(metadata_headers + [line])
            status = _split_status_line(
                line, column_lengths, self._cluster_name, job_metadata
            )
            jobs.append(status)
        return jobs

    def get_job(self, job_id: str) -> Optional[JobStatus]:
        """Gets the specified job's status.

        Args:
            job_id: The ID of the job to get.

        Returns:
            The job status if found, None otherwise.
        """
        job_list = self.list_jobs()
        for job in job_list:
            if job.id == job_id:
                return job
        return None

    def get_latest_job(self) -> Optional[JobStatus]:
        """Gets the most recent job on this cluster."""
        job_list = self.list_jobs()
        if len(job_list) == 0:
            return None
        return job_list[-1]

    def cancel(self, job_id) -> Optional[JobStatus]:
        """Cancels the specified job.

        Args:
            job_id: The ID of the job to cancel.

        Returns:
            The job status if found, None otherwise.
        """
        command = f"scancel {job_id}"
        result = self.run_commands([command])
        if result.exit_code != 0:
            raise RuntimeError(f"Failed to cancel job. stderr: {result.stderr}")
        return self.get_job(job_id)

    @retry_auth
    def put_recursive(self, source: str, destination: str) -> None:
        """Puts the specified file/directory to the remote path using rsync.

        Args:
            source: The local file/directory to write.
            destination: The remote path to write the file/directory to.
        """
        if Path(source).is_dir():
            self.run_commands([f"mkdir -p {destination}"])
        docs_dir = Path(source) / "docs"
        tests_dir = Path(source) / "tests"
        git_ignore = Path(source) / ".gitignore"
        rsync_cmd_list = [f'rsync -e "ssh {_CTRL_PATH}" -avz --delete ']
        if git_ignore.is_file():
            rsync_cmd_list.append(f"--exclude-from {str(git_ignore)} ")
        if tests_dir.is_dir():
            rsync_cmd_list.append(f"--exclude {str(tests_dir)} ")
        if docs_dir.is_dir():
            rsync_cmd_list.append(f"--exclude {str(docs_dir)} ")
        rsync_cmd_list.append(f"{source} ")
        rsync_cmd_list.append(f"{self._user}@{self._slurm_host}:{destination}")
        rsync_cmd = "".join(rsync_cmd_list)
        logger.info(f"Running rsync command: {rsync_cmd} ...")
        try:
            child = subprocess.run(
                rsync_cmd,
                shell=True,
                capture_output=True,
                timeout=300,
            )
            logger.info(f"Rsync command completed with exit code: {child.returncode}")
            if child.returncode != 0:
                parsed_stderr = child.stderr.decode("utf-8") if child.stderr else ""
                raise RuntimeError(f"Rsync failed. stderr: {parsed_stderr}")
        except subprocess.TimeoutExpired:
            raise RuntimeError("Timeout while running rsync command.")

    def put(self, file_contents: str, destination: str) -> None:
        """Puts the specified file contents to the remote path.

        Args:
            file_contents: The contents of the file to write.
            destination: The remote path to write the file to.
        """
        destination_path = Path(destination)
        parent_dir = destination_path.parent
        dir_cmd = f"mkdir -p {parent_dir}"
        create_cmd = f"touch {destination}"
        write_command = f'cat <<"SCRIPTFILETAG" > {destination}'
        file_suffix = "SCRIPTFILETAG"
        cmds = [dir_cmd, create_cmd, write_command, file_contents, file_suffix]
        result = self.run_commands(cmds)
        if result.exit_code != 0:
            raise RuntimeError(f"Failed to write file. stderr: {result.stderr}")

    def get_logs_stream(
        self, cluster_name: str, job_id: Optional[str] = None
    ) -> SlurmLogStream:
        """Gets a stream that tails the logs of the target job.

        Args:
            cluster_name: The name of the cluster the job was run in.
            job_id: The ID of the job to tail the logs of.
        """
        if job_id is None:
            lastest_job_id = self.get_latest_job()
            if lastest_job_id is None:
                raise RuntimeError("No jobs found on the cluster.")
            logger.info(
                f"No job ID provided. Using the most recent job ID: {lastest_job_id.id}"
            )
            job_id = lastest_job_id.id
        return SlurmLogStream(cluster_name, job_id, client=self)



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import hashlib
import logging
import os
import re
from typing import Optional

from oumi.core.tokenizers.base_tokenizer import BaseTokenizer


def sanitize_run_name(run_name: Optional[str]) -> Optional[str]:
    """Computes a sanitized version of wandb run name.

    A valid run name may only contain alphanumeric characters, dashes, underscores,
    and dots, with length not exceeding max limit.

    Args:
        run_name: The original raw value of run name.
    """
    if not run_name:
        return run_name

    # Technically, the limit is 128 chars, but we limit to 100 characters
    # because the system may generate aux artifact names e.g., by prepending a prefix
    # (e.g., "model-") to our original run name, which are also subject
    # to max 128 chars limit.
    _MAX_RUN_NAME_LENGTH = 100

    # Replace all unsupported characters with '_'.
    result = re.sub("[^a-zA-Z0-9\\_\\-\\.]", "_", run_name)
    if len(result) > _MAX_RUN_NAME_LENGTH:
        suffix = "..." + hashlib.shake_128(run_name.encode("utf-8")).hexdigest(8)
        result = result[0 : (_MAX_RUN_NAME_LENGTH - len(suffix))] + suffix

    if result != run_name:
        logger = logging.getLogger("oumi")
        logger.warning(f"Run name '{run_name}' got sanitized to '{result}'")
    return result


def try_str_to_bool(s: str) -> Optional[bool]:
    """Attempts to convert a string representation to a boolean value.

    This function interprets various string inputs as boolean values.
    It is case-insensitive and recognizes common boolean representations.

    Args:
        s: The string to convert to a boolean.

    Returns:
        bool: The boolean interpretation of the input string, or `None`
            for unrecognized string values.

    Examples:
        >>> str_to_bool("true") # doctest: +SKIP
        True
        >>> str_to_bool("FALSE") # doctest: +SKIP
        False
        >>> str_to_bool("1") # doctest: +SKIP
        True
        >>> str_to_bool("no") # doctest: +SKIP
        False
        >>> str_to_bool("peach") # doctest: +SKIP
        None
    """
    s = s.strip().lower()

    if s in ("true", "yes", "1", "on", "t", "y"):
        return True
    elif s in ("false", "no", "0", "off", "f", "n"):
        return False
    return None


def str_to_bool(s: str) -> bool:
    """Convert a string representation to a boolean value.

    This function interprets various string inputs as boolean values.
    It is case-insensitive and recognizes common boolean representations.

    Args:
        s: The string to convert to a boolean.

    Returns:
        bool: The boolean interpretation of the input string.

    Raises:
        ValueError: If the input string cannot be interpreted as a boolean.

    Examples:
        >>> str_to_bool("true") # doctest: +SKIP
        True
        >>> str_to_bool("FALSE") # doctest: +SKIP
        False
        >>> str_to_bool("1") # doctest: +SKIP
        True
        >>> str_to_bool("no") # doctest: +SKIP
        False
    """
    result = try_str_to_bool(s)

    if result is None:
        raise ValueError(f"Cannot convert '{s}' to boolean.")
    return result


def compute_utf8_len(s: str) -> int:
    """Computes string length in UTF-8 bytes."""
    # This is inefficient: allocates a temporary copy of string content.
    # FIXME Can we do better?
    return len(s.encode("utf-8"))


def get_editable_install_override_env_var() -> bool:
    """Returns whether OUMI_FORCE_EDITABLE_INSTALL env var is set to a truthy value."""
    s = os.environ.get("OUMI_FORCE_EDITABLE_INSTALL", "")
    mode = s.lower().strip()
    bool_result = try_str_to_bool(mode)
    if bool_result is not None:
        return bool_result
    return False


# Experimental function, only for developer usage.
def set_oumi_install_editable(setup: str) -> str:
    """Tries to replace oumi PyPi installs with editable installation from source.

    For example, the following line:
        `pip install uv && uv pip -q install oumi[gpu,dev] vllm`
    will be replaced with:
        `pip install uv && uv pip -q install -e '.[gpu,dev]' vllm`

    Args:
        setup (str): The bash setup script to modify. May be multi-line.

    Returns:
        The modified setup script.
    """
    setup_lines = setup.split("\n")
    for i, line in enumerate(setup_lines):
        # Skip comments.
        if line.strip().startswith("#"):
            continue

        # In summary, this regex looks for variants of `pip install oumi` and replaces
        # the oumi package with an editable install from the current directory.
        #
        # Tip: Use https://regexr.com/ or an LLM to help understand the regex.
        # It captures any misc. tokens like flags for the pip and
        # install commands, in addition to any optional dependencies oumi is installed
        # with.
        #
        # `((?:[-'\"\w]+ +)*)` matches whitespace-separated tokens potentially
        # containing quotes, such as flag names and values.
        # `((?:[-'\",\[\]\w]+ +)*)` does the same, with the addition of commas and
        # brackets, which may be present for packages with optional dependencies.
        # Since these don't include special characters like && and ;, it shouldn't span
        # across multiple pip install commands.
        # `(?<!-e )` means we don't match if the previous token is -e. This means an
        # editable install of a local dir called "oumi" is being done, so we skip it.
        # NOTE: We ideally should check for `--editable` as well, but Python re doesn't
        # support lookbehinds with variable length.
        # We additionally consume quotation marks around oumi if present.
        # Finally, `(\[[^\]]*\])?['\"]?` captures optional dependencies, if present.
        pattern = (
            r"pip3? +((?:[-'\"\w]+ +)*)install +((?:[-'\",\[\]\w]+ +)*)"
            r"(?<!-e )['\"]?oumi(\[[^\]]*\])?['\"]?"
        )
        # Compared to the pattern we captured, the changes are replacing `oumi` with
        # `.` and adding `-e` to make the install editable.
        replacement = r"pip \1install \2-e '.\3'"

        result = re.sub(pattern, replacement, line)
        if result == line:
            continue
        # Replace the line in the setup script.
        logger = logging.getLogger("oumi")
        logger.info(f"Detected the following oumi installation: `{line}`")
        logger.info(f"Replaced with: `{result}`")
        setup_lines[i] = result
    return "\n".join(setup_lines)


def truncate_to_max_tokens_limit(
    text: str,
    tokenizer: BaseTokenizer,
    *,
    max_tokens: int,
    truncation_side: str = "right",
) -> tuple[str, int]:
    """Truncates text to `max_length` in tokens.

    Args:
        text: A text prompt.
        tokenizer: The tokenizer used for encoding the data.
        max_tokens: Maximum number of tokens to keep.
        truncation_side: The side to truncate the tokens ("right" or "left").

    Returns:
        A tuple containing truncated text prompt and the number of tokens.
    """
    if max_tokens <= 0:
        raise ValueError("`max_tokens` must be a positive integer")
    elif truncation_side not in ("left", "right"):
        raise ValueError(
            f"Invalid truncation_side: '{truncation_side}'. Expected 'left' or 'right'."
        )

    if not text:
        return ("", 0)

    left_side = truncation_side == "left"

    # The `truncation_side` parameter isn't universally supported by all tokenizers.
    # Let's do left-side truncation as post-processing.
    result = tokenizer(
        text,
        return_offsets_mapping=True,
        return_length=True,
        max_length=(None if left_side else max_tokens),
        truncation=(not left_side),  # Left-side truncation is done as post-processing.
    )
    if "offset_mapping" not in result:
        raise RuntimeError(
            f"Tokenizer must return offset mapping for truncation! Got: {result.keys()}"
        )
    token2char_offsets = result["offset_mapping"]
    if not isinstance(token2char_offsets, list):
        raise RuntimeError(
            "offset_mapping returned by tokenizer is not a list! "
            f"Got: {type(token2char_offsets)}"
        )

    truncated_text: str = ""
    num_truncated_tokens: int = 0
    if len(token2char_offsets) > 0:
        num_truncated_tokens = min(len(token2char_offsets), max_tokens)
        if left_side:
            lead_token_start = token2char_offsets[-num_truncated_tokens][0]
            if not (lead_token_start >= 0 and lead_token_start < len(text)):
                raise RuntimeError(
                    f"Truncation error: lead_token_start={lead_token_start} "
                    f"for text of length {len(text)}"
                )
            truncated_text = text[lead_token_start:]
        else:
            last_token_end = token2char_offsets[num_truncated_tokens - 1][1]
            if not (last_token_end >= 0 and last_token_end <= len(text)):
                raise RuntimeError(
                    f"Truncation error: last_token_end={last_token_end} "
                    f"for text of length {len(text)}"
                )
            truncated_text = text[:last_token_end]

    return (truncated_text, num_truncated_tokens)


def truncate_text_pieces_to_max_tokens_limit(
    text_pieces: list[str],
    tokenizer: BaseTokenizer,
    *,
    max_tokens: int,
    truncation_side: str = "right",
) -> list[str]:
    """Truncates text pieces to total length not exceeding `max_length`.

    Args:
        text_pieces: A list of text prompts.
        tokenizer: The tokenizer used for encoding the data.
        max_tokens: Maximum number of tokens to keep in all text pieces combined.
        truncation_side: The side to truncate the tokens ("right" or "left").

    Returns:
        A list of truncated text prompts.
    """
    if max_tokens <= 0:
        raise ValueError("`max_tokens` must be a positive integer")
    remaining_tokens = max_tokens

    result = copy.deepcopy(text_pieces)
    if truncation_side == "left":
        result.reverse()

    for idx, text_piece in enumerate(result):
        if len(text_piece) == 0:
            continue
        elif remaining_tokens > 0:
            truncated_text_piece, num_tokens = truncate_to_max_tokens_limit(
                text_piece,
                tokenizer=tokenizer,
                max_tokens=remaining_tokens,
                truncation_side=truncation_side,
            )
            result[idx] = truncated_text_piece
            remaining_tokens -= num_tokens
        else:
            result[idx] = ""

    if truncation_side == "left":
        result.reverse()

    return result



import math
import time

import torch

from oumi.performance.telemetry import (
    CudaTimerContext,
    TelemetryTracker,
    TimerContext,
)
from tests.markers import requires_gpus


#
# Timer
#
def test_timer_context():
    measurements = []
    with TimerContext("test_timer", measurements):
        time.sleep(0.1)

    assert len(measurements) == 1
    assert math.isclose(0.1, measurements[0], rel_tol=0.1)


def test_timer_context_as_decorator():
    measurements = []

    @TimerContext("test_decorator", measurements)
    def sample_function():
        time.sleep(0.1)

    sample_function()

    assert len(measurements) == 1
    assert math.isclose(0.1, measurements[0], rel_tol=0.1)


#
# Cuda Timer
#
@requires_gpus()
def test_cuda_timer_context():
    measurements = []
    with CudaTimerContext("test_cuda_timer", measurements):
        time.sleep(0.1)

    assert len(measurements) == 1
    assert measurements[0] > 0


@requires_gpus()
def test_cuda_timer_context_as_decorator():
    measurements = []

    @CudaTimerContext("test_cuda_decorator", measurements)
    def sample_cuda_function():
        time.sleep(0.1)

    sample_cuda_function()

    assert len(measurements) == 1
    assert measurements[0] > 0


#
# Telemetry Tracker
#
def test_telemetry_tracker_timer():
    tracker = TelemetryTracker()

    with tracker.timer("test_operation"):
        time.sleep(0.1)

    summary = tracker.get_summary()
    assert "test_operation" in summary["timers"]
    assert math.isclose(0.1, summary["timers"]["test_operation"]["total"], rel_tol=0.1)

    tracker.print_summary()


@requires_gpus()
def test_telemetry_tracker_cuda_timer():
    tracker = TelemetryTracker()

    with tracker.cuda_timer("test_cuda_operation"):
        torch.cuda.synchronize()
        time.sleep(0.1)

    summary = tracker.get_summary()
    assert "test_cuda_operation" in summary["cuda_timers"]
    assert summary["cuda_timers"]["test_cuda_operation"]["mean"] > 0

    tracker.print_summary()


@requires_gpus()
def test_telemetry_tracker_log_gpu_memory():
    tracker = TelemetryTracker()

    tracker.log_gpu_memory()

    summary = tracker.get_summary()
    assert len(summary["gpu_memory"]) == 1
    assert "allocated" in summary["gpu_memory"][0]
    assert "reserved" in summary["gpu_memory"][0]

    tracker.print_summary()


@requires_gpus()
def test_telemetry_tracker_record_gpu_temperature():
    tracker = TelemetryTracker()

    tracker.record_gpu_temperature()
    tracker.record_gpu_temperature()
    tracker.record_gpu_temperature()

    summary = tracker.get_summary()
    assert "gpu_temperature" in summary
    assert len(summary["gpu_temperature"]) == 6
    assert summary["gpu_temperature"]["count"] == 3

    assert (
        summary["gpu_temperature"]["min"] > 0
        and summary["gpu_temperature"]["min"] < 100
    )
    assert (
        summary["gpu_temperature"]["max"] > 0
        and summary["gpu_temperature"]["max"] < 100
    )
    assert (
        summary["gpu_temperature"]["mean"] > 0
        and summary["gpu_temperature"]["mean"] < 100
    )
    assert (
        summary["gpu_temperature"]["median"] > 0
        and summary["gpu_temperature"]["median"] < 100
    )
    assert summary["gpu_temperature"]["std_dev"] >= 0

    tracker.print_summary()


def test_telemetry_tracker_get_summary():
    tracker = TelemetryTracker()

    with tracker.timer("operation1"):
        time.sleep(0.1)

    with tracker.timer("operation2"):
        time.sleep(0.2)

    summary = tracker.get_summary()
    assert "total_time" in summary
    assert "timers" in summary
    assert "operation1" in summary["timers"]
    assert "operation2" in summary["timers"]
    assert (
        summary["timers"]["operation2"]["total"]
        > summary["timers"]["operation1"]["total"]
    )

    all_summaries = tracker.get_summaries_from_all_ranks()
    assert len(all_summaries) == 1
    assert "total_time" in all_summaries[0]
    assert "timers" in all_summaries[0]
    assert "operation1" in all_summaries[0]["timers"]
    assert "operation2" in all_summaries[0]["timers"]
    assert (
        all_summaries[0]["timers"]["operation2"]["total"]
        > all_summaries[0]["timers"]["operation1"]["total"]
    )

    info = tracker.compute_cross_rank_summaries(
        all_summaries, measurement_names={"total_time"}
    )
    assert "total_time" in info
    assert len(info) == 1

    info = info["total_time"]
    assert len(info) == 8
    assert "count" in info and info["count"] == 1.0
    assert "max" in info and info["max"] > 0.0
    assert "max_index" in info and info["max_index"] == 0
    assert "mean" in info and info["mean"] > 0.0 and info["mean"] == info["max"]
    assert "median" in info and info["median"] > 0.0 and info["median"] == info["max"]
    assert "min" in info and info["min"] > 0.0 and info["min"] == info["max"]
    assert "min_index" in info and info["min_index"] == 0
    assert "std_dev" in info and info["std_dev"] == 0

    info = tracker.compute_cross_rank_summaries(
        all_summaries,
        measurement_names={
            "timers": {
                "operation1": {"median"},
                "operation2": {"total", "max"},
            },
        },
    )
    assert "timers" in info
    assert isinstance(info["timers"], dict)
    assert len(info["timers"]) == 2

    assert "operation1" in info["timers"]
    assert isinstance(info["timers"]["operation1"], dict)
    assert len(info["timers"]["operation1"]) == 1
    assert "median" in info["timers"]["operation1"]
    assert isinstance(info["timers"]["operation1"]["median"], dict)
    assert len(info["timers"]["operation1"]["median"]) == 8

    assert "operation2" in info["timers"]
    assert "total" in info["timers"]["operation2"]
    assert isinstance(info["timers"]["operation2"]["total"], dict)
    assert len(info["timers"]["operation2"]["total"]) == 8
    assert "max" in info["timers"]["operation2"]
    assert isinstance(info["timers"]["operation2"]["max"], dict)
    assert len(info["timers"]["operation2"]["max"]) == 8


@requires_gpus()
def test_telemetry_tracker_get_summary_with_gpu_temperature():
    tracker = TelemetryTracker()

    with tracker.timer("operation1"):
        time.sleep(0.1)

    with tracker.timer("operation2"):
        time.sleep(0.2)

    tracker.record_gpu_temperature()
    tracker.record_gpu_temperature()
    tracker.record_gpu_temperature()

    summary = tracker.get_summary()
    assert "total_time" in summary
    assert "gpu_temperature" in summary
    assert len(summary["gpu_temperature"]) == 6
    assert "timers" in summary
    assert "operation1" in summary["timers"]
    assert "operation2" in summary["timers"]
    assert (
        summary["timers"]["operation2"]["total"]
        > summary["timers"]["operation1"]["total"]
    )

    all_summaries = tracker.get_summaries_from_all_ranks()
    assert len(all_summaries) == 1
    assert "total_time" in all_summaries[0]
    assert "gpu_temperature" in all_summaries[0]
    assert "timers" in all_summaries[0]
    assert "operation1" in all_summaries[0]["timers"]
    assert "operation2" in all_summaries[0]["timers"]
    assert (
        all_summaries[0]["timers"]["operation2"]["total"]
        > all_summaries[0]["timers"]["operation1"]["total"]
    )

    info = tracker.compute_cross_rank_summaries(
        all_summaries, measurement_names={"total_time"}
    )
    assert "total_time" in info
    assert len(info) == 1

    info = info["total_time"]
    assert len(info) == 8
    assert "count" in info and info["count"] == 1.0
    assert "max" in info and info["max"] > 0.0
    assert "max_index" in info and info["max_index"] == 0
    assert "mean" in info and info["mean"] > 0.0 and info["mean"] == info["max"]
    assert "median" in info and info["median"] > 0.0 and info["median"] == info["max"]
    assert "min" in info and info["min"] > 0.0 and info["min"] == info["max"]
    assert "min_index" in info and info["min_index"] == 0
    assert "std_dev" in info and info["std_dev"] == 0

    info = tracker.compute_cross_rank_summaries(
        all_summaries,
        measurement_names={
            "timers": {
                "operation1": {"median"},
                "operation2": {"total", "max"},
            },
            "gpu_temperature": {"max"},
        },
    )
    assert "timers" in info
    assert isinstance(info["timers"], dict)
    assert len(info["timers"]) == 2
    assert "gpu_temperature" in info
    assert isinstance(info["gpu_temperature"], dict)
    assert len(info["gpu_temperature"]) == 1

    assert "max" in info["gpu_temperature"]
    assert isinstance(info["gpu_temperature"]["max"], dict)
    assert len(info["gpu_temperature"]["max"]) == 8

    assert "count" in info["gpu_temperature"]["max"]
    assert "mean" in info["gpu_temperature"]["max"]
    assert "median" in info["gpu_temperature"]["max"]
    assert "std_dev" in info["gpu_temperature"]["max"]
    assert "min" in info["gpu_temperature"]["max"]
    assert "min_index" in info["gpu_temperature"]["max"]
    assert "max" in info["gpu_temperature"]["max"]
    assert "max_index" in info["gpu_temperature"]["max"]

    assert "operation1" in info["timers"]
    assert isinstance(info["timers"]["operation1"], dict)
    assert len(info["timers"]["operation1"]) == 1
    assert "median" in info["timers"]["operation1"]
    assert isinstance(info["timers"]["operation1"]["median"], dict)
    assert len(info["timers"]["operation1"]["median"]) == 8

    assert "operation2" in info["timers"]
    assert "total" in info["timers"]["operation2"]
    assert isinstance(info["timers"]["operation2"]["total"], dict)
    assert len(info["timers"]["operation2"]["total"]) == 8
    assert "max" in info["timers"]["operation2"]
    assert isinstance(info["timers"]["operation2"]["max"], dict)
    assert len(info["timers"]["operation2"]["max"]) == 8



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from dataclasses import dataclass, field, fields
from typing import Any, Optional

from oumi.core.configs.params.base_params import BaseParams


@dataclass
class GrpoParams(BaseParams):
    model_init_kwargs: dict[str, Any] = field(default_factory=dict)
    """Keyword arguments for `AutoModelForCausalLM.from_pretrained(...)`"""

    max_prompt_length: Optional[int] = None
    """Maximum length of the prompt.

    If the prompt is longer than this value, it will be truncated left.
    If unspecified (`None`), defaults to 512.
    """

    max_completion_length: Optional[int] = None
    """Maximum length of the generated completion.

    If unspecified (`None`), defaults to 256.
    """

    num_generations: Optional[int] = None
    """Number of generations per prompt to sample.

    The global batch size (num_processes * per_device_batch_size) must be divisible
    by this value. If unspecified (`None`), defaults to 8.
    """

    temperature: float = 0.9
    """Temperature for sampling.

    The higher the temperature, the more random the completions.
    """

    remove_unused_columns: bool = False
    """Whether to only keep the column `"prompt"` in the dataset.

    If you use a custom reward function that requires any column other than `"prompts"`
    and `"completions"`, you should set it to `False`.
    """

    repetition_penalty: Optional[float] = 1.0
    """Float that penalizes new tokens if they appear in the prompt/response so far.

    Values > 1.0 encourage the model to use new tokens, while values < 1.0 encourage
    the model to repeat tokens.
    """

    use_vllm: bool = False
    """Whether to use vLLM for generating completions.

    If set to `True`, ensure that a GPU is kept unused for training,
    as vLLM will require one for generation.
    """

    vllm_mode: Optional[str] = None
    """The mode to use for vLLM generation ("colocate" or "server").

    If set to `None`, defaults to "server".

    Server mode means that vLLM is running on a
    separate server that the trainer will communicate with. It requires the server to
    be started with `trl vllm-serve` beforehand.

    Colocate mode means that vLLM will run in the same process as the trainer and share
    GPUs. While this is simpler as it doesn't require a separate server, vLLM will
    contend with the trainer for GPU resources.
    """

    vllm_gpu_memory_utilization: float = 0.9
    """Ratio (between 0 and 1) of GPU memory to reserve.

    Fraction of VRAM reserved  for the model weights, activations, and KV cache on
    the device dedicated to generation powered by vLLM. Higher values will increase
    the KV cache size and thus improve the model's throughput.
    However, if the value is too high, it may cause out-of-memory (OOM) errors
    during initialization.
    """

    epsilon: float = 0.2
    """Epsilon value for clipping the relative probability in the loss.

    For example, if epsilon is 0.2, then the new probability can only differ from
    the old probability by a factor of x0.8-1.2."""

    log_completions: bool = False
    """Whether to log prompt and completion pairs every `logging_steps` steps."""

    rollout_function: Optional[str] = None
    """The name of the rollout function in the Oumi registry to use for generation in
    trl's GRPO trainer. This field is ignored if not using trl' GRPO trainer. If not
    specified, the default rollout function will be used."""

    def __post_init__(self):
        """Verifies params."""
        if self.max_prompt_length is not None and self.max_prompt_length <= 0:
            raise ValueError(
                "GrpoParams.max_prompt_length must be positive. "
                f"Actual: {self.max_prompt_length}"
            )
        if self.max_completion_length is not None and self.max_completion_length <= 0:
            raise ValueError(
                "GrpoParams.max_completion_length must be positive. "
                f"Actual: {self.max_completion_length}"
            )
        if self.num_generations is not None and self.num_generations <= 0:
            raise ValueError(
                "GrpoParams.num_generations must be positive. "
                f"Actual: {self.num_generations}"
            )
        if not (
            math.isfinite(self.temperature)
            and self.temperature >= 0.0
            and self.temperature <= 1.0
        ):
            raise ValueError(
                "GrpoParams.temperature must be within [0.0, 1.0] range. "
                f"Actual: {self.temperature}"
            )

    def to_hf_trainer_kwargs(self) -> dict[str, Any]:
        """Converts GrpoParams to TRL's GRPOConfig kwargs."""
        result = {}
        if len(self.model_init_kwargs) > 0:
            result["model_init_kwargs"] = self.model_init_kwargs
        if self.max_prompt_length is not None:
            result["max_prompt_length"] = self.max_prompt_length
        if self.max_completion_length is not None:
            result["max_completion_length"] = self.max_completion_length
        if self.num_generations is not None:
            result["num_generations"] = self.num_generations

        already_processed_keys: set[str] = set(
            {
                "model_init_kwargs",
                "max_prompt_length",
                "max_completion_length",
                "num_generations",
                "rollout_function",  # Will be passed into GRPO trainer.
            }
        )

        # Copy the majority of fields that aren't special-cased.
        for param in fields(self):
            if param.name.startswith("vllm_") or param.name in already_processed_keys:
                continue
            result[param.name] = getattr(self, param.name)

        if self.use_vllm:  # Return vLLM params only if vLLM is enabled.
            if self.vllm_mode is not None:
                result["vllm_mode"] = self.vllm_mode
            result["vllm_gpu_memory_utilization"] = self.vllm_gpu_memory_utilization
        return result
